
from dataclasses import dataclass
import textwrap
from typing import cast
from relationalai.clients.snowflake import Provider

@dataclass
class Procedure:
    provider: Provider
    database: str
    schema: str
    name: str

    def create(self, python_code: str, args: list[tuple[str, str]] = [], returns: str = "STRING"):
        arg_string = ", ".join([f"{arg[0]} {arg[1]}" for arg in args])
        sql_query = textwrap.dedent(f"""
        CREATE OR REPLACE PROCEDURE {self.database}.{self.schema}.{self.name}({arg_string})
        RETURNS {returns}
        LANGUAGE PYTHON
        RUNTIME_VERSION = '3.11'
        PACKAGES = ('snowflake-snowpark-python')
        HANDLER = 'main'
        EXECUTE AS CALLER
        AS
        $$
        {textwrap.indent(python_code, " " * 8)}
        $$;
        """)
        return self.provider.sql(sql_query)
    
    def execute(self, *args):
        arg_string = ", ".join([f"'{arg}'" if isinstance(arg, str) else str(arg) for arg in args])
        sql_query = f"CALL {self.database}.{self.schema}.{self.name}({arg_string});"
        return self.provider.sql(sql_query)
    
    def delete(self):
        for row in cast(list, self.provider.sql(f"show procedures in {self.database}.{self.schema};")):
            if row["name"].upper() == self.name.upper():
                name_with_input_args = row["arguments"].split(" RETURN ")[0]
                self.provider.sql(f"drop procedure {self.database}.{self.schema}.{name_with_input_args}")

    def get_python_code(self, database, schema, sproc_name):
        query = f"""
            SELECT PROCEDURE_DEFINITION
            FROM {database.upper()}.INFORMATION_SCHEMA.PROCEDURES
            WHERE PROCEDURE_NAME = '{sproc_name.upper()}'
            AND PROCEDURE_SCHEMA = '{schema.upper()}';
        """
        response = cast(list, self.provider.sql(query))

        if not response:
            raise ValueError(f"Stored procedure '{sproc_name}' not found in {database}.{schema}.")

        procedure_definition = response[0]['PROCEDURE_DEFINITION']
        if not procedure_definition:
            raise ValueError(f"No definition found for stored procedure '{sproc_name}'.")

        return procedure_definition



from abc import abstractmethod

from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.rules import Rule

def addrule(rul): return contextStack.root_context().addrule(rul)


# A @rule decorator that can be used to register a rule, optionally with one or more
# annotations.
#
# Examples:
#   @rule(Annotation.INNER_LOOP_NON_STRATIFIED, ...)
#   def my_rule():
#       ...
#
#   @rule()
#   def my_other_rule():
#       ...
def rule(*annotations):
    def decorator(func):
        rule_instance = addrule(func)
        for annotation in annotations:
            rule_instance.annotate(annotation)
        return func

    return decorator


class Relation:
    class NullContainer:
        def name(self): return ""

        def entity(self): return False

    # A Relation can serve as a context for declaring rules that
    # populate that relation and also for declaring constraints
    # that pertain only to this single relation -- e.g., functional
    # and mandatory constraints.
    #
    def __enter__(self):
        contextStack.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()
        pass

    # A Relation can be "invoked" in the body of a rule definition to
    # declare what will be an Atom in the Schema generated by the rule
    # def.
    #
    # Note that because Schemas expect these Atoms to be defined on
    # RelationComponents, we have to make sure that such a component
    # exists in the Schema in context.
    #
    def __call__(self, *args):
        context = contextStack.root_context()
        if context is not None:
            return context.build_atom(self, args)
        return None

    def __init__(self, namespace, name, sig):
        self._relname = name
        self._container = namespace
        self._signature = sig
        self._rules = []

    # The "rule" to add is a function object with a number of parameters that should
    # equal the number of roles in this Relation.
    #
    def addrule(self, rule):
        if not isinstance(rule, Rule):
            # Assume that rule is a Python function that, when invoked can build a rule
            rule = Rule(self).elaborate(rule)

        self._rules.append(rule)
        return rule

    # Add a sequence of reading text snippets to use when verbalizing
    # this Relation.
    #
    def addverb(self, verb):
        self._verb = verb
        return self

    def arity(self):
        return self._signature.arity()

    def display(self):
        return self.qualified_name()

    def entityid(self):
        return hash(self.qualified_name())

    def function(self):
        self._signature.declare_functional()

    # Generates a new Rule that populates this relation, returning a tuple whose
    # first component is the Rule object and whose next n-1 components are the
    # free variables of this rule in order. Allows one to construct new rules that
    # populate a relation and get a handle on all free variables individually using
    # one method call, e.g.:
    #
    #   (rule, x, y, z) = relation.genrule()
    #
    def genrule(self):
        rule = Rule(self)
        self._rules.append(rule)
        tup = [rule]
        for v in rule.freevars():
            tup.append(v)
        return tuple(tup)

    def pprint(self):
        return self.display() + self._signature.display()

    # Returns the fully-qualified name that will be used in queries to
    # refer to this Relation.
    #
    def qualified_name(self):
        if not isinstance(self._container, Namespace):
            raise Exception(f"Cannot find containing namespace for Relation named {self._relname}")

        if self._container == Namespace.top:
            return self._relname
        else:
            # TODO: [VAMI] We changed `::` to `__` to spead up E2E testing. We should change it back to `::` before release.
            return f"{self._container.qualified_name()}__{self._relname}"

class AbstractRelationSignature:

    def __init__(self, *types):
        self._types = types

    def __call__(self, *args, **kwargs):
        seq = [self]
        for i in range(len(args)):
            seq.append(args[i])
        context = contextStack.root_context()
        if context is not None:
            return context.build_relation_variable(seq, kwargs)
        return None

    def arity(self):
        return len(self._types)

    def entityid(self):
        relhash = hash([r.entityid() for r in self._types])
        if self.functional():
            return hash((relhash, True))
        else:
            return hash((relhash, False))

    def relational(self) -> bool:
        return True

    # An n-ary AbstractRelationSignature is functional if its n-th column
    # is functionally dependent on the first n-1 columns.
    #
    @abstractmethod
    def functional(self) -> bool:
        """Return True when relation is functional otherwise False."""

    def domainsig(self):
        if self.functional():
            a = self.arity()
            return self._types[0:a - 1]
        else:
            return self._types

    def rangesig(self):
        if self.functional():
            return self._types[self.arity() - 1]
        else:
            return None

    def display(self):
        dsp = [r.display() for r in self.domainsig() if r is not None]
        dom = "(" + ", ".join(dsp) + ")"
        range_sig = self.rangesig()
        if range_sig is not None:
            return f"{dom} -> {range_sig.display()}"
        else:
            return dom

    # An AbstractRelationSignature is a subtype of another if the two have the
    # same number of types and in each corresponding type pair (t1, t2),
    # the type that plays t1 is a subtype of the type that plays t2.
    #
    def subtype_of(self, sig2):
        if self.arity() != sig2.arity():
            return False

        for i in range(self.arity()):
            if not self._types[i].subtype_of(sig2._types[i]):
                return False

        return True

class RelationSignature(AbstractRelationSignature):

    def __init__(self, *types):
        super().__init__(*types)
        self._is_functional = False

    def declare_functional(self):
        self._is_functional = True

    def functional(self): return self._is_functional



from __future__ import annotations

from pandas import DataFrame
from typing import Any, Optional, Union
import relationalai as rai

from relationalai.early_access.metamodel import ir, compiler, executor as e
from relationalai.early_access.rel import Compiler

class RelExecutor(e.Executor):
    """Executes Rel code using the RAI client."""

    def __init__(self, database: str, dry_run: bool = False) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run


    def execute(self, model: ir.Model, observer: Optional[compiler.Observer]=None) -> Union[DataFrame, Any]:
        """
        Execute Rel code for a given model/database name.

        Args:
            model: The name of the model/database
            rel_code: The fragment to execute
            rules: Optional list of rule fragments to compile before the rel_code

        Returns:
            The result of executing the Rel code
        """

        rel_compiler = Compiler()
        full_code = rel_compiler.compile(model, observer)

        # from logscope import log
        # log(full_code)

        if self.dry_run:
            return DataFrame()

        resources = rai.clients.snowflake.Resources()
        resources.config.set("use_graph_index", False)

        try:
            resources.create_graph(self.database)
        except Exception as e:
            if "already exists" not in str(e).lower():
                raise e

        engine = resources.config.get("engine")
        return resources.exec_raw(self.database, engine, full_code)



from relationalai.early_access.dsl.core.builders.logic import LogicBuilder
from relationalai.early_access.dsl.core.constraints.predicate.universal import RelationalEqualityConstraint
from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.exprs.scalar import ScalarExprBuilder
from relationalai.early_access.dsl.core.types.constrained import ConstrainedValueType
from relationalai.early_access.dsl.constants import PRIME, DASHES
from relationalai.early_access.dsl.schemas.components import RelationalComponent, ScalarComponent
from relationalai.early_access.dsl.types.entities import AbstractEntityType


def remove_prefix(nm, prefix):
    lpref = len(prefix)
    if lpref == 0 or not nm.startswith(prefix):
        return nm
    else:
        truncated = nm[lpref:]
        if len(truncated) == 1:
            return truncated.lower()
        else:
            newstr = truncated[0].lower() + truncated[1:]
            return newstr


def extract_name(nm, dashes, prefix):
    dash_suffix = dashes * PRIME
    index = nm.rfind(dash_suffix)
    if index == -1:
        return remove_prefix(nm, prefix)
    else:
        return remove_prefix(nm[0:index], prefix)


def properly_decorated_name(nm, dashes, prefix):
    if dashes != 0:
        dash_suffix = dashes * PRIME
        if nm.rfind(dash_suffix) == -1:
            return False
    return nm.startswith(prefix)


class SchemaContext(LogicBuilder, ScalarExprBuilder):

    @staticmethod
    def close():
        contextStack.pop()

    @staticmethod
    def open(schema):
        ctx = SchemaContext(schema)
        contextStack.push(ctx)
        return ctx

    def __init__(self, schema):
        self.schema = schema

    def build_atom(self, rel, args):
        s = self.schema
        atom = s.build_atom(rel, args)
        s.add_atomic_constraint(atom)
        return atom

    def build_comparison(self, left, op, right):
        s = self.schema
        if left.relational():
            if not right.relational() or op != "=":
                bad_expr = f"{left.display()} {op} {right.display()}"
                raise Exception(
                    f"Currently cannot declare anything other than equality constraints on relations. Saw {bad_expr}")
            c = RelationalEqualityConstraint(left, right)
            s.add_predicate_constraint(c)
            return c
        else:
            c = s.build_comparison(left, op, right)
            s.add_scalar_constraint(c)
            return c

    def build_element_of(self, e, set):
        s = self.schema
        s.add_atomic_constraint(s.build_element_of(e, set))

    def build_scalar_variable(self, args, kwargs):

        dashes = kwargs[DASHES] if DASHES in kwargs else 0

        tp = args[0]
        nameorvar = args[1]
        if isinstance(nameorvar, str):
            name = nameorvar

            if self.schema.disjunctive():
                raise Exception("Cannot add a component directly to a disjunctive schema")

            decl = ScalarComponent(name, tp, dashes)
            self.schema.add_component(decl)
            return decl
        else:
            if isinstance(tp, AbstractEntityType):
                rel = tp.population()
                relname = rel.qualified_name()
                ctx = self.schema
                if ctx.has_relation_component(relname):
                    comp = ctx.retrieve_relation_component(relname)
                else:
                    # Then add a new component for this relation to the
                    # schema that is under construction
                    #
                    # todo: figure out if we use `build_scalar_variable` at all
                    comp = rel.signature(relname) # type: ignore

                # Now assert the atom
                comp(nameorvar)
            else:
                raise Exception(f"Illegal to range over the population of infinite type {tp.display()}")

    def build_relation_variable(self, args, kwargs):

        dashes = kwargs[DASHES] if DASHES in kwargs else 0

        tp = args[0]

        # [REKS] Very subtle. This is needed because when tp is a ConstrainedValueType,
        #        this method will be called. Correct during type analysis, but incorrect
        #        during Schema declaration
        #
        if isinstance(tp, ConstrainedValueType):
            return self.build_scalar_variable(args, kwargs)

        nameorvar = args[1]
        if isinstance(nameorvar, str):
            name = nameorvar

            if self.schema.disjunctive():
                raise Exception("Cannot add a component directly to a disjunctive schema")

            decl = RelationalComponent(name, tp, dashes)

            self.schema.add_relational_component(decl)
            return decl
        else:
            if isinstance(tp, AbstractEntityType):
                rel = tp.population()
                relname = rel.qualified_name()
                ctx = self.schema
                if ctx.has_relation_component(relname):
                    comp = ctx.retrieve_relation_component(relname)
                else:
                    # Then add a new component for this relation to the
                    # schema that is under construction
                    #
                    # todo: figure out if we use `build_relation_variable` at all
                    comp = rel.signature(relname) # type: ignore

                # Now assert the atom
                comp(nameorvar)
            else:
                raise Exception(f"Illegal to range over the population of infinite type {tp.display()}")



from relationalai.early_access.dsl.core.types.constrained.subtype import ValueSubtype
from relationalai.early_access.dsl.core.types.constrained.nominal import ValueType
from relationalai.early_access.dsl.core.types.unconstrained import UnconstrainedValueType, UnconstrainedNumericType

standard_value_types = {}

# Hash types
Hash = UnconstrainedValueType('Hash')
standard_value_types[Hash.name()] = Hash

# Boolean types
Boolean = UnconstrainedNumericType('Boolean')
standard_value_types[Boolean.name()] = Boolean

# Decimal types
#
Decimal = UnconstrainedNumericType('Decimal')
standard_value_types[Decimal.name()] = Decimal

UnsignedDecimal = ValueSubtype('UnsignedDecimal')
standard_value_types[UnsignedDecimal.name()] = UnsignedDecimal
with UnsignedDecimal:
    x = Decimal()
    x >= 0

PositiveDecimal = ValueSubtype('PositiveDecimal')
standard_value_types[PositiveDecimal.name()] = PositiveDecimal
with PositiveDecimal:
    x = UnsignedDecimal()
    x != 0

# Floating-point types
#
Float = UnconstrainedNumericType('Float')
standard_value_types[Float.name()] = Float
Double = UnconstrainedNumericType('Double')  # A float64 in Rel
standard_value_types[Double.name()] = Double

# Integer types
#
BigInteger = UnconstrainedNumericType('BigInteger')  # 128-bit integer
standard_value_types[BigInteger.name()] = BigInteger

BigUnsignedInteger = ValueSubtype('BigUnsignedInteger')
standard_value_types[BigUnsignedInteger.name()] = BigUnsignedInteger
with BigUnsignedInteger:
    x = BigInteger()
    x >= 0

BigPositiveInteger = ValueSubtype('BigPositiveInteger')
standard_value_types[BigPositiveInteger.name()] = BigPositiveInteger
with BigPositiveInteger:
    x = BigUnsignedInteger()
    x != 0

Integer = UnconstrainedNumericType('Integer')
standard_value_types[Integer.name()] = Integer

UnsignedInteger = ValueSubtype('UnsignedInteger')
standard_value_types[UnsignedInteger.name()] = UnsignedInteger
with UnsignedInteger:
    x = Integer()
    x >= 0

PositiveInteger = ValueSubtype('PositiveInteger')
standard_value_types[PositiveInteger.name()] = PositiveInteger
with PositiveInteger:
    x = UnsignedInteger()
    x != 0

# Date types
#
Date = ValueType('Date')
standard_value_types[Date.name()] = Date
with Date:
    s = Integer()

# DateTime types
#
DateTime = ValueType('DateTime')
standard_value_types[DateTime.name()] = DateTime
with DateTime:
    s = Integer()

RowId = ValueType('RowId', BigInteger)

# String types
#
String = UnconstrainedValueType('String')
standard_value_types[String.name()] = String

Any = UnconstrainedValueType('Any')
standard_value_types[Any.name()] = Any

Symbol = UnconstrainedValueType('Symbol')
standard_value_types[Symbol.name()] = Symbol



from abc import abstractmethod

from relationalai.early_access.dsl.core.constraints.scalar import entails
from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.types.variables import TypeVariable
from relationalai.early_access.dsl.core.utils import generate_stable_uuid


# This abstract class generalizes over all types
#
class Type:

    # A Type T can be "instantiated" for one of several purposes, depending
    # on context:
    #
    #   1) in the context of a ConstrainedValueType, to add a parameter TypeVariable
    #      to that type; and
    #   2) in the context of a ConjunctiveSchema:
    #      a) if the argument is a string name, then adds a new ScalarComponent of
    #         that name to the schema; or
    #      b) if the argument is a ScalarVariable, then adds an atom that references
    #         the unary population relation for the EntityType in a Rel body formula
    #
    # In case (1), there should be no arguments (args); in case (2a), we expect there
    # to be exactly one string argument that names the component being added to the
    # SchemaFragment; while in context (2b) we expect there to be exactly one argument
    # of type ScalarVariable.
    #
    # Because we use RelationVariables to range over the parameters of a nominal value
    # type, to properly handle case (1) we have to call one of:
    #
    #  - build_relation_variable
    #  - build_scalar_variable
    #
    # depending on whether T is a nominal type. But that distinction is really only
    # relevant to case (1) -- which means the concrete builder objects that implement
    # the ScalarConstraintBuilder interface for cases (2a) and (2b) must implement
    # both methods the same way.
    #
    def __call__(self, *args, **kwargs):
        seq = [self]
        for i in range(len(args)):
            seq.append(args[i])

        if self.nominal() or self.entity():
            return contextStack.root_context().build_relation_variable(seq, kwargs)
        else:
            return contextStack.root_context().build_scalar_variable(seq, kwargs)

    def __enter__(self):
        contextStack.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()
        pass

    def __init__(self, nm):
        self._name = nm
        #
        if not contextStack.empty():
            tp = contextStack.root_context()
            if isinstance(tp, Namespace):
                self._namespace = tp
            else:
                self._namespace = Namespace.top
        else:
            self._namespace = Namespace.top

    # Returns the number of TypeVariables that parameterize this Type
    #
    def arity(self) -> int:
        return 1

    def display(self):
        return self.name()

    # Returns True if this type is an EntityType or one of its subtypes
    # and False otherwise
    #
    def entity(self) -> bool:
        return False

    # Returns an integer (hash) that identifies this Type, allowing us
    # to treat Types sort of like entity types for which we had declared
    # a preferred identifier to allow different physical objects to unify
    # to the same entity. Contrast this with the "id" method which would
    # not allow us to do this unification.
    #
    def entityid(self):
        return hash(self.name())

    def guid(self):
        return generate_stable_uuid(self._name)

    # Returns the Rel required to instantiate an instance of this Type
    # from a given value val. By default, just return val; subclasses
    # will override as appropriate
    #
    def instantiate(self, *val):
        return val[0]

    def name(self):
        return self._name

    def namespace(self):
        return self._namespace

    # This predicate returns True if this type is a ValueType or
    # some subtype of a ValueType.
    #
    def nominal(self) -> bool:
        return False


class AbstractValueType(Type):

    def __init__(self, nm):
        Type.__init__(self, nm)

    # Returns true if self is both unconstrained and also numeric, meaning its constraints
    # (if any) can be statically verified.
    #
    def basic_numeric(self) -> bool: return False

    # Returns True if the Python constant `val` is an instance of this ValueType
    def contains(self, val) -> bool: return False

    # Gathers the set of BasicTypeVariable constraints that this AbstractValueType denotes
    # either by declaring or inheriting.
    #
    def denotes(self): return {}

    # Returns true if this is a nominal value type or a subtype that is rooted by a nominal
    # value type
    #
    def nominal(self) -> bool: return False

    # Returns true if this is a constraint value type or a subtype that is rooted by a constrained
    # value type
    #
    def constraint(self) -> bool: return False

    # Returns true if this is a constraint value type or a subtype that is rooted by a constrained
    # value type and signature_arity is 1
    #
    def constraint_unary(self) -> bool:
        return self.constraint() and self.unary()

    # Creates a new TypeVariable that ranges over this AbstractValueType, including all
    # constraints (if any) suitably instantiated to the new TypeVariable
    #
    @abstractmethod
    def ranges_over(self) -> TypeVariable: pass

    # Where possible, returns the lone UnconstrainedValueType that will be used as the physical
    # representation of instances of this AbstractValueType at runtime.
    #
    def root_nominal_type(self):
        raise Exception(f"Cannot determine root nominal type of AbstractValueType {self.name()}")

    # The signature arity of a AbstractValueType refers to the number of BasicTypeVariables
    # that occur in its physical implementation. For UnconstrainedValueTypes, the arity is 1.
    # For ConstrainedValueTypes, the arity is the size of the flattened basic-type variable
    # tuple that it constrains.
    #
    def signature_arity(self) -> int: return 1

    # Returns true if this is a value type has signature_arity == 1
    #
    def unary(self):
        return self.signature_arity() == 1

    def subtype_of(self, T2):
        if self.nominal():
            if not T2.nominal():
                return False

            # Both are nominal value types, so make sure names are the same
            #
            nt1 = self.root_nominal_type()
            nt2 = T2.root_nominal_type()
            if nt1.name != nt2.name:
                return False
        else:
            if T2.nominal():
                return False

            # Neither type is nominal, so check if one or both are unconstrained

            if self.unconstrained():
                return self.name == T2.name

            if T2.unconstrained():
                u2 = self.root_unconstrained_type() # type: ignore
                return T2.name == u2.name

        # All nominal type checking has passed and neither type is Unconstrained

        if self.signature_arity() != T2.signature_arity():
            return False

        # Types have same signature arities, so check constraints on type variables.

        mapping = map_type_var_tuple_to(self.typevar_tuple(), T2.typevar_tuple())

        if len(mapping) != self.signature_arity():
            return False

        # Same typevar tuples, so now just check that the every models of self is also
        # a model of T2.
        #
        cset1 = {}
        for c in self.denotes().values():
            cprime = c.rename(mapping)
            cset1[cprime.entityid()] = cprime

        cset2 = T2.denotes()

        # [[cset1]] NOT subseteq [[cset2]] if there exists a c in cset2 that is not entailed
        # by cset1
        #
        for c in cset2:
            if not entails(cset1, cset2[c]):
                return False

        return True

    # Returns true if this type is Unconstrained
    #
    def unconstrained(self) -> bool:
        return False

    @abstractmethod
    def typevar_tuple(self) -> tuple[TypeVariable, ...]:
        pass


def basic_type_instance(val):
    if isinstance(val, str):
        return True
    if isinstance(val, int):
        return True
    if isinstance(val, float):
        return True
    return False

def map_type_var_tuple_to(tup1, tup2):
    return map_type_var_tuple_to_rec(tup1, tup2, {})

def map_type_var_tuple_to_rec(tup1, tup2, m):
    if type(tup1) is tuple:
        if type(tup2) is not tuple:
            return {}

        v1 = tup1[0]
        v2 = tup2[0]
        m[v1.display()] = v2
        return map_type_var_tuple_to_rec(tup1[1:], tup2[1:], m)
    else:
        if type(tup2) is tuple:
            return {}
        m[tup1.display()] = tup2

    return m

def typevar_tuple_string_rec(tup) -> str:
    if type(tup) is not tuple:
        return tup.display()

    results = [typevar_tuple_string_rec(v) for v in tup]

    if len(results) == 1:
        return results[0]
    else:
        return "(" + ", ".join(results) + ")"


def typevar_tuple_string(tup) -> str:
    if type(tup) is not tuple:
        return f"({tup.display()})"
    else:
        return typevar_tuple_string_rec(tup)



from __future__ import annotations
import json
from typing import Any, List, Set, cast
from datetime import datetime, date
import pandas as pd
from zoneinfo import ZoneInfo

from relationalai.errors import RAIException
from relationalai.rel_utils import RESERVED_WORDS, sanitize_identifier

from .metamodel import Builtins, Action, ActionType, Namer, Property, Type, Var, Task, Value
from . import metamodel as m
from . import compiler as c
import re

gather_vars = m.Utils.gather_vars

#--------------------------------------------------
# Emitter
#--------------------------------------------------

rel_infix = [">", "<", ">=", "<=", "=", "!=", "+", "-", "*", "/", "^", "~="]
rel_sanitize_re = re.compile(r'[^\w:\[\]\^" ,]|^(?=\d)', re.UNICODE)
unsafe_symbol_pattern = re.compile(r"[^a-zA-Z0-9_]", re.UNICODE)

def sanitize(input_string, is_rel_name_or_symbol = False):
    # Replace non-alphanumeric characters with underscores
    if is_rel_name_or_symbol and "[" in input_string:
        string_parts = input_string.split('[', 1)
        sanitized_name_or_symbol = sanitize_identifier(string_parts[0])
        sanitized_rest = re.sub(rel_sanitize_re, "_", string_parts[1])
        sanitized = f"{sanitized_name_or_symbol}[{sanitized_rest}"
    else:
        if "::" in input_string: # TODO: This is a temp solution to avoid sanitizing the namespace
            sanitized = re.sub(rel_sanitize_re, "_", input_string)
        else:
            sanitized = sanitize_identifier(input_string)

    # Check if the resulting string is a keyword and append an underscore if it is
    if sanitized in RESERVED_WORDS:
        sanitized += "_"

    return sanitized

class Emitter(c.Emitter):

    def __init__(self, config):
        super().__init__()
        self.namer = Namer()
        self.stack:List[Set[Var]] = []
        self.mapped = {}
        self.config = config

    def reset(self):
        self.namer.reset()
        self.stack = []
        self.mapped = {}

    #--------------------------------------------------
    # Emit
    #--------------------------------------------------

    def emit(self, task: Task|Var):
        self.mapped.clear()
        code = ""
        try:
            if isinstance(task, Task):
                code = getattr(self, task.behavior.value)(task)
            elif isinstance(task, Var):
                code = self.emit_var(task)
        except Exception as e:
            # print("EMIT FAILED:", e)
            raise e
        return code

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def sanitize(self, input_string, is_rel_name = False):
        return sanitize(input_string, is_rel_name)

    #--------------------------------------------------
    # Vars
    #--------------------------------------------------

    def emit_val(self, value: Value, var:Var|None=None):
        if value is True:
            return 'boolean_true'
        if value is False:
            return 'boolean_false'
        if isinstance(value, list):
            return f"{', '.join([self.emit_var(v) for v in value])}"
        if isinstance(value, Task) and (value.items or not value.name):
            return self.to_inline_relation(value)
        if isinstance(value, Task) and value.name:
            return f"{{{value.name}}}"
        if isinstance(value, bytes):
            if var and var.type.isa(Builtins.ValueType):
                return f"^{self.sanitize(var.name).capitalize()}Type[0x{value.hex()}]"
            return f"0x{value.hex()}"
        if isinstance(value, Property):
            return "{" + self.sanitize(value.name) + "}"
        if isinstance(value, Type):
            return "{" + self.sanitize(value.name, value.isa(Builtins.Relation)) + "}"
        if isinstance(value, pd.Timestamp):
            t = value.tz_localize('UTC')
            return t.isoformat()
        if isinstance(value, datetime):
            t = value.astimezone(ZoneInfo('UTC'))
            return t.isoformat()
        if isinstance(value, date):
            t = value
            return t.isoformat()
        if isinstance(value, str):
            # % is the string interpolation character in rel
            return json.dumps(value).replace("%", "\\%")
        return json.dumps(value)

    def emit_var(self, var: Var|Value):
        if not isinstance(var, Var):
            return self.emit_val(var)
        if var in self.mapped:
            return self.mapped[var]
        if var.type.isa(Builtins.Symbol):
            if unsafe_symbol_pattern.search(cast(str, var.value)):
                return f":\"{var.value}\""
            return f":{self.sanitize(var.value)}"
        if var.value is not None:
            return self.emit_val(var.value, var)
        if var.name == "_":
            return "_"
        name = self.namer.get(var)
        return f"_{self.sanitize(name)}"

    def to_name(self, value: Value|Var|None) -> str:
        if value is not None and hasattr(value, "name"):
            name = getattr(value, "name") # @NOTE: Working around pyright
            if name:
                return name

        raise RAIException(f"Unable to determine the appropriate name for value of type: {type(value)} ({value})")

    #--------------------------------------------------
    # Sequence
    #--------------------------------------------------

    def sequence_action(self, action: Action):
        if action.entity.value == Builtins.RawCode:
            return str([*action.bindings.values()][0].value)
        elif isinstance(action.entity.value, Task):
            return self.emit(action.entity.value)
        else:
            raise Exception(f"TODO: Rel emit for action type {action.action}")

    def sequence(self, task: Task):
        items = [ self.sequence_action(i) for i in task.items ]
        return "\n\n".join(items)

    #--------------------------------------------------
    # Query helpers
    #--------------------------------------------------

    def to_relation(self, action: Action, vars: set, body:List[str], in_head = False):
        root = cast(Type, action.entity.value)
        rel_name = root.name
        if not rel_name and isinstance(root, Task):
            rel_name = f"T{root.id}"
        elif not rel_name:
            rel_name = f"R_{root.id}"
        args = []
        for var in action.bindings.values():
            emitted = self.emit_var(var)
            vars.add(var)
            if in_head and emitted in args:
                name = self.namer.get(var)
                orig = emitted
                emitted = self.sanitize(f"_{self.namer.get_safe_name(name)}")
                body.append(f"{emitted} = {orig}")
            elif in_head and var.value is not None and var.isa(Builtins.ValueType):
                name = self.namer.get(var)
                emitted = self.sanitize(f"_{self.namer.get_safe_name(name)}")
                body.append(f"{emitted} = {self.emit_val(var.value, var)}")
            elif in_head and isinstance(var.value, bool):
                name = self.namer.get(var)
                orig = emitted
                emitted = self.sanitize(f"_{self.namer.get_safe_name(name)}")
                body.append(f"{emitted} = {orig}")
            args.append(emitted)
        if rel_name in rel_infix:
            if len(args) == 2:
                return f"{args[0]} {rel_name} {args[1]}"
            else:
                return f"{args[2]} = {args[0]} {rel_name} {args[1]}"
        rel_name = self.sanitize(rel_name, True)
        if isinstance(root, Task) and root.isa(Builtins.Quantifier):
            # TODO: handle quantifier grouping
            args = args[1:]
            rel_name = rel_name.lower()
            if root.isa(Builtins.Exists):
                rel_name = "not empty"
        if action.action == ActionType.Persist:
            rel_name = f"insert[:{rel_name}]"
        elif action.action == ActionType.Unpersist:
            rel_name = f"delete[:{rel_name}]"
        partial_appl = rel_name.find('[')
        if partial_appl >= 0:
            final_rel_name = rel_name[:partial_appl]
            if len(args):
                final_args = f"{rel_name[partial_appl+1:-1]}, {', '.join(args)}"
            else:
                final_args = ', '.join(args)
            final = f"{final_rel_name}({final_args})"
        else:
            final = f"{rel_name}({', '.join(args)})"
        return final

    def to_return(self, action: Action, vars: set, body:List[str]):
        args = []
        for var in action.bindings.values():
            emitted = self.emit_var(var)
            vars.add(var)
            if emitted in args:
                name = self.namer.get(var)
                orig = emitted
                emitted = self.sanitize(f"_{self.namer.get_safe_name(name)}")
                body.append(f"{emitted} = {orig}")
            elif var.value is not None and var.isa(Builtins.ValueType):
                name = self.namer.get(var)
                emitted = self.sanitize(f"_{self.namer.get_safe_name(name)}")
                body.append(f"{emitted} = {self.emit_val(var.value, var)}")
            elif isinstance(var.value, bool):
                name = self.namer.get(var)
                orig = emitted
                emitted = self.sanitize(f"_{self.namer.get_safe_name(name)}")
                body.append(f"{emitted} = {orig}")
            args.append(emitted)
        return f"output({', '.join(args)})"

    def to_set(self, action: Action, vars: set):
        vals:Any = [v for v in action.bindings.values()]
        mid_point = len(vals) // 2
        vars_str = ", ".join([self.emit_var(i) for i in vals[mid_point:]])
        rows = []
        for ix in range(len(vals[0].value)):
            row_vals = vals[:mid_point]
            product_str = ",".join([self.emit_var(col.value[ix]) for col in row_vals])
            if len(row_vals) > 1:
                rows.append(f"({product_str})")
            else:
                rows.append(product_str)
        row_str = "; ".join(rows)
        return f"{{{row_str}}}({vars_str})"

    def to_inline_set(self, action: Action, vars: set):
        vals:Any = [v for v in action.bindings.values()]
        rows = []
        for ix in range(len(vals[0].value)):
            row_vals = vals[:-1]
            product_str = ",".join([self.emit_var(col.value[ix]) for col in row_vals])
            if len(row_vals) > 1:
                rows.append(f"({product_str})")
            else:
                rows.append(product_str)
        row_str = "; ".join(rows)
        return (vals[-1], f"{{{row_str}}}")

    def to_inline_relation(self, task:Task, existing_vars=set()):
        return f"{{{self.query(task, True)}}}"

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, task: Task, inline = False):
        supporting_rules = []
        head = ""
        body = []
        body_vars = set()
        head_vars = set()

        task_vars = set(gather_vars(task.items))
        self.stack.append(task_vars)

        for i in task.items:
            if i.action in [ActionType.Get, ActionType.Call]:
                if i.entity.value == Builtins.RawData:
                    body.append(self.to_set(i, body_vars))
                elif i.entity.value == Builtins.InlineRawData:
                    (v, data) = self.to_inline_set(i, body_vars)
                    self.mapped[v] = data
                    head_vars.add(v)
                elif i.entity.value == Builtins.Install:
                    (item,) = i.bindings.values()
                    supporting_rules.append(f"declare {self.sanitize(self.to_name(item.value), True)}")
                    if isinstance(item.value, Type) and item.value.isa(Builtins.ValueType):
                        vtype = self.sanitize(item.value.name, True).capitalize() + "Type"
                        supporting_rules.append(f"value type {vtype} {{ UInt128 }}")
                        supporting_rules.append(f"@inline def pyrel_unwrap(y in {vtype}, x): ^{vtype}(x, y)")
                        supporting_rules.append(f"def ::std::common::uuid_string(x in {vtype}, y): exists((z) | pyrel_unwrap(x, z) and uuid_string(z, y))")
                        for op in ["<", ">"]:
                            supporting_rules.append(f"def ::std::common::({op})[x in {vtype}, y in {vtype}]: exists((a, b) | ^{vtype}(a, x) and ^{vtype}(b, y) and a {op} b)")
                            if self.config.get("compiler.use_monotype_operators", False):
                                supporting_rules.append(f"def ::std::monotype::({op})[x in {vtype}, y in {vtype}]: exists((a, b) | ^{vtype}(a, x) and ^{vtype}(b, y) and a {op} b)")
                        for op in ["minimum", "maximum"]:
                            supporting_rules.append(f"def ::std::common::{op}(x in {vtype}, y in {vtype}, z in {vtype}): exists((a, b) | ^{vtype}(minimum[a, b], z) and ^{vtype}(a, x) and ^{vtype}(b, y))")
                else:
                    body.append(self.to_relation(i, body_vars, body))
            elif i.action == ActionType.Bind and i.entity.value == Builtins.Return:
                head_rel = self.to_return(i, head_vars, body)
                head = f"def {head_rel}:\n    "
            elif i.action in [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]:
                if not inline:
                    head_rel = self.to_relation(i, head_vars, body, True)
                    annotations = set(["from_pyrel"])
                    # If optimized rel is requested, add desired attribute
                    opt = self.config.get(name="describe_optimization", strict=False)
                    if opt == "simple" :
                        annotations.add(Builtins.InspectOptimizedSimple.name.lower())
                    elif opt == "verbose" :
                        annotations.add(Builtins.InspectOptimizedVerbose.name.lower())
                    ent = i.entity.value
                    if isinstance(ent, Type) or isinstance(ent, Property):
                        for parent in ent.parents:
                            if parent.isa(Builtins.Annotation):
                                annotations.add(parent.name.lower())
                    if annotations:
                        annotation_str = " ".join([f"@{a}" for a in annotations])
                        head = f"{annotation_str}\ndef {head_rel}:\n    "
                    else:
                        head = f"def {head_rel}:\n    "
                else:
                    props = list(i.bindings.values())
                    head_vars.update(props)
                    head = f"({', '.join([self.emit_var(v) for v in props])}): "
            elif i.action == ActionType.Construct:
                inputs = [*i.bindings.values()]
                input_str =', '.join([self.emit_var(v) for v in inputs[:-1]])
                output = self.emit_var(inputs[-1])
                body_vars.update(inputs)
                body.append(f"{output} = ^{self.sanitize(self.to_name(i.entity.value)).capitalize()}Type[{input_str}]")
            else:
                raise Exception(f"TODO: Rel emit for action type {i.action}")
        body_str = " and\n    ".join(body)
        existing_vars = set()
        for vars in self.stack[:-1]:
            existing_vars.update(vars)
        from_vars = [self.emit_var(v) for v in (body_vars - head_vars - existing_vars) if v.value is None]
        if len(from_vars) and not inline:
            body_str = f"exists(({', '.join(from_vars)}) |\n    {body_str})"
        if head and len(from_vars) and inline:
            body_str = f"exists(({', '.join(from_vars)}) | {body_str})"


        self.stack.pop()

        if not head and inline:
            return body_str
        elif not head:
            head = f"def T{task.id}():\n    "

        support_str = ""
        if len(supporting_rules):
            support_str = "\n\n".join(supporting_rules) + "\n\n"

        if support_str and not head_vars and not body_str:
            return support_str

        if not body_str:
            body_str = "{()}"


        return f"{support_str}{head}{body_str}"



"""
    Intermediate Representation of RelationalAI programs.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from itertools import count
from typing import Any, IO, Optional, Tuple, TypeVar, Union as PyUnion

from .util import FrozenOrderedSet

import json

#--------------------------------------------------
# IR Nodes
#--------------------------------------------------

_global_id = count(0)
def next_id():
    return next(_global_id)

@dataclass(frozen=True)
class Node:
    # A generated id that is not used on comparisons and hashes
    id: int = field(default_factory=next_id, init=False, compare=False, hash=False)

    @property
    def kind(self):
        return self.__class__.__name__.lower()

    def __str__(self):
        return node_to_string(self)

#-------------------------------------------------
# Public Types - Model
#-------------------------------------------------

@dataclass(frozen=True)
class Model(Node):
    """Represents the whole universe of elements that make up a program."""
    engines: FrozenOrderedSet["Engine"]
    relations: FrozenOrderedSet["Relation"]
    types: FrozenOrderedSet["Type"]
    root: Task


#-------------------------------------------------
# Public Types - Engine
#-------------------------------------------------

@dataclass(frozen=True)
class Capability(Node):
    """Engine capabilities, such as 'graph algorithms', 'solver', 'constant time count', etc"""
    name: str

@dataclass(frozen=True)
class Engine(Node):
    """The entity that owns a Task and provides access to certain relations."""
    name: str
    platform: str # SQL, Rel, JS, OpenAI, etc
    info: Any
    capabilities: FrozenOrderedSet[Capability]
    relations: FrozenOrderedSet["Relation"]


#-------------------------------------------------
# Public Types - Data Model
#-------------------------------------------------

@dataclass(frozen=True)
class ScalarType(Node):
    """The named type."""
    name: str

@dataclass(frozen=True)
class ListType(Node):
    """A type that represents a list of elements of some other type."""
    element_type: Type

@dataclass(frozen=True)
class SetType(Node):
    """A type that represents a set of elements of some other type."""
    element_type: Type

@dataclass(frozen=True)
class UnionType(Node):
    """A type that represents either one of a set of types."""
    types: FrozenOrderedSet[Type]

# The type of a field in a relation
Type = PyUnion[ScalarType, ListType, SetType, UnionType]

@dataclass(frozen=True)
class Field(Node):
    """A named field in a relation."""
    name: str
    type: Type
    input: bool # must be grounded as the relation cannot compute it


@dataclass(frozen=True)
class Relation(Node):
    """A relation represents the schema of a set of tuples."""
    name: str
    fields: Tuple[Field, ...]
    requires: FrozenOrderedSet[Capability]


#-------------------------------------------------
# Public Types - Tasks
#-------------------------------------------------

@dataclass(frozen=True)
class Task(Node):
    engine: Optional[Engine]

#
# Task composition
#

@dataclass(frozen=True)
class Logical(Task):
    """Execute sub-tasks up to fix-point."""
    # Executes tasks concurrently. Succeeds if every task succeeds.
    hoisted: Tuple[VarOrElse, ...]
    body: Tuple[Task, ...]

@dataclass(frozen=True)
class Union(Task):
    """Execute sub-tasks in any order."""
    # Executes tasks concurrently. Succeeds if at least one task succeeds.
    hoisted: Tuple[VarOrElse, ...]
    tasks: Tuple[Task, ...]

@dataclass(frozen=True)
class Sequence(Task):
    """Execute sub-tasks one at a time, in this order."""
    # Executes tasks in order. Stops when a task fails. Succeeds if all tasks succeed.
    hoisted: Tuple[VarOrElse, ...]
    tasks: Tuple[Task, ...]

@dataclass(frozen=True)
class Match(Task):
    """Execute sub-tasks in order until the first succeeds."""
    # Executes tasks in order. Stops when a task succeeds. Succeeds if some task succeeds.
    hoisted: Tuple[VarOrElse, ...]
    tasks: Tuple[Task, ...]

@dataclass(frozen=True)
class Until(Task):
    """Execute both `check` and `body` concurrently, until check succeeds."""
    hoisted: Tuple[VarOrElse, ...]
    check: Task
    body: Task

@dataclass(frozen=True)
class Wait(Task):
    hoisted: Tuple[VarOrElse, ...]
    check: Task

# TODO: DynamicLookup


#
# Logical Quantifiers
#

@dataclass(frozen=True)
class Not(Task):
    """Logical negation of the sub-task."""
    task: Task

@dataclass(frozen=True)
class Exists(Task):
    """Existential quantification over the sub-task."""
    vars: Tuple[Var, ...]
    task: Task

@dataclass(frozen=True)
class ForAll(Task):
    """Universal quantification over the sub-task."""
    vars: Tuple[Var, ...]
    task: Task


#
# Iteration (Loops)
#

# loops body until a break condition is met
@dataclass(frozen=True)
class Loop(Task):
    """Execute the body in a loop, incrementing the iter variable, until a break sub-task in
    the body succeeds."""
    hoisted: Tuple[VarOrElse, ...]
    iter: Var
    body: Task

@dataclass(frozen=True)
class Break(Task):
    """Break a surrounding Loop if the `check` succeeds."""
    check: Task


#
# Relational Operations
#

@dataclass(frozen=True)
class Var(Node):
    """A named variable that can point to objects of this type."""
    type: Type
    name: str

@dataclass(frozen=True)
class Else(Node):
    """A variable with a default value. This can be used in 'hoisted' attributes to
    represent the value to assign a variable if the underlying task fails."""
    var: Var
    value: Value

VarOrElse = PyUnion[Var, Else]

@dataclass(frozen=True)
class Literal(Node):
    """A literal value with a specific type."""
    type: Type
    value: Any

Value = PyUnion[str, int, float, bool, None, Var, Else, Literal, Type, Relation, Tuple["Value", ...], FrozenOrderedSet["Value"]]

@dataclass(frozen=True)
class Annotation(Node):
    """Meta information that can be attached to Updates."""
    relation: Relation
    args: Tuple[Value, ...]

class Effect(Enum):
    """Possible effects of an Update."""
    derive = "derive"
    insert = "insert"
    delete = "delete"

@dataclass(frozen=True)
class Update(Task):
    """Updates the relation with these arguments. The update can derive new tuples
    temporarily, can insert new tuples persistently, or delete previously persisted tuples."""
    relation: Relation
    args: Tuple[Value, ...]
    effect: Effect
    annotations: FrozenOrderedSet[Annotation]

@dataclass(frozen=True)
class Lookup(Task):
    """Lookup tuples from this relation, filtering with these arguments."""
    relation: Relation
    args: Tuple[Value, ...]

@dataclass(frozen=True)
class Output(Task):
    """Output the value of these vars, giving them these column names."""
    aliases: FrozenOrderedSet[Tuple[str, Var]]

@dataclass(frozen=True)
class Construct(Task):
    """Construct an id from these values, and bind the id to this var."""
    values: Tuple[Value, ...]
    id_var: Var

@dataclass(frozen=True)
class Aggregate(Task):
    """Perform an aggregation with these arguments."""
    aggregation: Relation
    projection: Tuple[Var, ...]
    group: Tuple[Var, ...]
    args: Tuple[Value, ...]


#--------------------------------------------------
# Printer
#--------------------------------------------------

infix = ["+", "-", "*", "/", "%", "=", "!=", "<", "<=", ">", ">="]

T = TypeVar('T', bound=Node)
def node_to_string(node:Node|Tuple[T, ...]|FrozenOrderedSet[T]) -> str:
    io = StringIO()
    printer = Printer(io)
    printer.pprint(node)
    return io.getvalue()

def value_to_string(value:PyUnion[Value, Tuple[Value, ...]]) -> str:
    return Printer().value_to_string(value)

@dataclass(frozen=True)
class Printer():
    io: Optional[IO[str]] = None

    # count of vars with this name, to generate names with ids when there's collision
    var_count: dict[str, int] = field(default_factory=dict)
    # cache of precomputed names for the var with this id
    var_name_cache: dict[int, str] =  field(default_factory=dict)

    def indent_print(self, depth, *args) -> None:
        """ Helper to print the arguments into the io with indented based on depth. """
        if self.io is None:
            print("    " * depth + " ".join(map(str, args)))
        else:
            self.io.write("    " * depth + " ".join(map(str, args)) + "\n")

    def print_hoisted(self, depth, name, hoisted: Tuple[VarOrElse, ...]):
        if hoisted:
            self.indent_print(depth, f"{name} ⇑[{', '.join([self.value_to_string(h) for h in hoisted])}]")
        else:
            self.indent_print(depth, name)

    def var_to_string(self, var: Var) -> str:
        if var.id in self.var_name_cache:
            return self.var_name_cache[var.id]

        if var.name in self.var_count:
            c = self.var_count[var.name]
            self.var_count[var.name] = c + 1
            name = f"{var.name}_{c}"
        else:
            self.var_count[var.name] = 2
            name = f"{var.name}"

        self.var_name_cache[var.id] = name
        return name

    def value_to_string(self, value:PyUnion[Value, Tuple[Value, ...]]) -> str:
        """ Return a string representation of the value. """
        if isinstance(value, (int, str, float, bool)):
            return json.dumps(value)
        elif value is None:
            return "None"
        elif isinstance(value, Var):
            return self.var_to_string(value)
        elif isinstance(value, Else):
            return f"{self.var_to_string(value.var)}={value.value}"
        elif isinstance(value, Literal):
            return f"{json.dumps(value.value)}"
        elif isinstance(value, ListType):
            return f"[{self.value_to_string(value.element_type)}]"
        elif isinstance(value, SetType):
            return f"{{{self.value_to_string(value.element_type)}}}"
        elif isinstance(value, UnionType):
            return f"{{{'; '.join(map(self.value_to_string, value.types))}}}"
        elif isinstance(value, ScalarType):
            return f"{value.name}"
        elif isinstance(value, Relation):
            return value.name
        elif isinstance(value, Tuple):
            return f"({', '.join(map(self.value_to_string, value))})"
        elif isinstance(value, FrozenOrderedSet):
            return f"{{{', '.join(map(self.value_to_string, value))}}}"
        else:
            raise NotImplementedError(f"value_to_string not implemented for {type(value)}")

    T = TypeVar('T', bound=Node)
    def pprint(self, node:Node|Tuple[T, ...]|FrozenOrderedSet[T], depth=0) -> None:
        """ Pretty print the node into the io, starting with indentation based on depth. If io is None,
        print into the standard output. """

        if isinstance(node, Tuple) or isinstance(node, FrozenOrderedSet):
            for n in node:
                self.pprint(n, depth)
        # model
        elif isinstance(node, Model):
            self.indent_print(depth, "Model")
            if len(node.engines) > 0:
                self.indent_print(depth + 1, "engines:")
            self.pprint(node.engines, depth + 2)
            if len(node.relations) > 0:
                self.indent_print(depth + 1, "relations:")
            self.pprint(node.relations, depth + 2)
            if len(node.types) > 0:
                self.indent_print(depth + 1, "types:")
            self.pprint(node.types, depth + 2)
            self.indent_print(depth + 1, "root:")
            self.pprint(node.root, depth + 2)

        # engine
        elif isinstance(node, Capability):
            self.indent_print(depth, node.name)
        elif isinstance(node, Engine):
            self.indent_print(depth, f"Engine ({node.name}, {node.platform})")
            self.indent_print(depth + 1, node.info)
            self.indent_print(depth + 1, ', '.join([c.name for c in node.capabilities]))
            self.pprint(node.relations, depth + 1)

        # data model
        elif isinstance(node, (ScalarType, ListType, SetType, UnionType)):
            self.indent_print(depth, self.value_to_string(node))
        elif isinstance(node, Field):
            s = f"{node.name}: {self.value_to_string(node.type)} {'(input)' if node.input else ''}"
            self.indent_print(depth, s)
        elif isinstance(node, Relation):
            self.indent_print(depth, node.name)
            self.pprint(node.fields, depth + 1)
            if len(node.requires) > 0:
                self.indent_print(depth + 1, "requires:")
                self.pprint(node.requires, depth + 2)

        # tasks

        # Task composition
        elif isinstance(node, Logical):
            self.print_hoisted(depth, "Logical", node.hoisted)
            self.pprint(node.body, depth + 1)
        elif isinstance(node, Sequence):
            self.print_hoisted(depth, "Sequence", node.hoisted)
            self.pprint(node.tasks, depth + 1)
        elif isinstance(node, Union):
            self.print_hoisted(depth, "Union", node.hoisted)
            self.pprint(node.tasks, depth + 1)
        elif isinstance(node, Match):
            self.print_hoisted(depth, "Match", node.hoisted)
            self.pprint(node.tasks, depth + 1)
        elif isinstance(node, Until):
            self.print_hoisted(depth, "Until", node.hoisted)
            self.pprint(node.check, depth + 1)
            self.pprint(node.body, depth + 1)
        elif isinstance(node, Wait):
            self.print_hoisted(depth, "Match", node.hoisted)
            self.pprint(node.check, depth + 1)

        # Relational Operations
        elif isinstance(node, Var):
            self.indent_print(0, self.value_to_string(node))
        elif isinstance(node, Literal):
            self.indent_print(0, self.value_to_string(node))
        elif isinstance(node, Annotation):
            if node.args:
                self.indent_print(depth, f"@{node.relation.name}{self.value_to_string(node.args)}")
            else:
                self.indent_print(depth, f"@{node.relation.name}")
        elif isinstance(node, Update):
            rel_name = node.relation.name
            annos = "" if not node.annotations else f" {' '.join(str(a) for a in node.annotations)}"
            self.indent_print(depth, f"→ {node.effect.value} {rel_name}{self.value_to_string(node.args)}{annos}")
        elif isinstance(node, Lookup):
            rel_name = node.relation.name
            if rel_name in infix:
                args = [self.value_to_string(arg) for arg in node.args]
                if len(node.args) == 2:
                    self.indent_print(depth, f"{args[0]} {rel_name} {args[1]}")
                elif len(node.args) == 1:
                    self.indent_print(depth, f"{rel_name}{args[0]}")
                elif len(node.args) == 3:
                    self.indent_print(depth, f"{args[2]} = {args[0]} {rel_name} {args[1]}")
            else:
                self.indent_print(depth, f"{rel_name}{self.value_to_string(node.args)}")
        elif isinstance(node, Output):
            args = []
            for k, v in node.aliases:
                ppv = self.value_to_string(v)
                if k != ppv:
                    args.append(f"{ppv} as '{k}'")
                else:
                    args.append(ppv)
            self.indent_print(depth, f"→ output({', '.join(args)})")
        elif isinstance(node, Construct):
            values = [self.value_to_string(v) for v in node.values]
            self.indent_print(depth, f"construct({', '.join(values)}, {self.value_to_string(node.id_var)})")
        elif isinstance(node, Aggregate):
            self.indent_print(depth, f"{node.aggregation.name}([{self.value_to_string(node.projection)}], [{self.value_to_string(node.group)}], [{self.value_to_string(node.args)}])")

        # Logical Quantifiers
        elif isinstance(node, Not):
            self.indent_print(depth, "Not")
            self.pprint(node.task, depth + 1)
        elif isinstance(node, Exists):
            self.indent_print(depth, f"Exists({', '.join([self.value_to_string(v) for v in node.vars])})")
            self.pprint(node.task, depth + 1)
        elif isinstance(node, ForAll):
            self.indent_print(depth, f"ForAll({', '.join([self.value_to_string(v) for v in node.vars])})")
            self.pprint(node.task, depth + 1)

        # Iteration (Loops)
        elif isinstance(node, Loop):
            self.print_hoisted(depth, f"Loop ⇓[{self.value_to_string(node.iter)}]", node.hoisted)
            self.pprint(node.body, depth + 1)

        elif isinstance(node, Break):
            self.indent_print(depth, "Break")
            self.pprint(node.check, depth + 1)

        elif isinstance(node, Task):
            # empty task represents success
            self.indent_print(depth, "Success")

        else:
            # return
            raise NotImplementedError(f"pprint not implemented for {type(node)}")



from typing import List, Tuple

import networkx as nx
from networkx.classes import DiGraph


def topological_sort(nodes: List[str], edges: List[Tuple[str, str]]) -> List[str]:
    # Check for cycles
    acyclic, dependency_graph =_is_acyclic_graph(nodes, edges)
    if not acyclic:
        raise ValueError("The graph contains a cycle")
    return list(nx.topological_sort(dependency_graph))

def is_acyclic_graph(nodes: List[str], edges: List[Tuple[str, str]]) -> bool:
    return _is_acyclic_graph(nodes, edges)[0]

def _is_acyclic_graph(nodes, edges) -> tuple[bool, DiGraph]:
    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return nx.is_directed_acyclic_graph(graph), graph


import pandas as pd

from relationalai.early_access.dsl.core.relations import rule, Relation
from relationalai.early_access.dsl.physical_metadata.tabular_source import TabularSource

class TabularSourceModule:
    def generate(self, table: TabularSource, data: pd.DataFrame):
        for index, row in data.iterrows():
            for column in data.columns:
                value = row[column]
                if pd.notna(value):
                    self._row_to_value_rule(table.__getattribute__(column), index, value)

    @staticmethod
    def _row_to_value_rule(relation: Relation, row, value):
        with relation:
            @rule()
            def row_to_value(r, v):
                r == row
                v == value


# type: ignore
import dataclasses
from typing import Optional

from dataclasses_json import LetterCase, dataclass_json


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class BoundRole:
    relationship: str
    player: str
    entityMapRef: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class BoundAttribute:
    name: str
    source: str
    hasValue: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class DataTransformerApplication:
    transformer: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class BindingObject:
    id: str
    attribute: Optional[BoundAttribute]
    role: BoundRole
    transformations: list[DataTransformerApplication] = dataclasses.field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class DataTransformer:
    name: str
    sourceType: str
    targetType: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Pattern:
    conceptName: str
    pattern: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class EntityMapRef:
    name: str
    source: str
    entityType: str
    relation: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class BindingModelObject:
    bindings: list[BindingObject] = dataclasses.field(default_factory=list)
    patterns: list[Pattern] = dataclasses.field(default_factory=list)
    dataTransformers: list[DataTransformer] = dataclasses.field(default_factory=list)
    entityMaps: list[EntityMapRef] = dataclasses.field(default_factory=list)



from enum import Enum

from relationalai.early_access.dsl.core import tempvar
from relationalai.early_access.dsl.core.builders.logic import LogicBuilder
from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.exprs.scalar import ScalarExprBuilder
from relationalai.early_access.dsl.core.logic import LogicFragment, RelVariable
from relationalai.early_access.dsl.core.logic.exists import ExistentialConstraint
from relationalai.early_access.dsl.core.types import Type


class Annotation(Enum):
    FUNCTION = 1
    INLINE = 2
    ITER = 3
    INNER_LOOP = 4
    INNER_LOOP_NON_STRATIFIED = 5

    def display(self):
        return f"@{self}"

    def __str__(self):
        return self.name.lower()


class Rule(LogicFragment, ScalarExprBuilder):

    def __enter__(self):
        contextStack.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()

    # Partially construct this Rule by creating a head atom and an array of free variables
    # from a Relation (rel)
    #
    def __init__(self, rel):
        LogicFragment.__init__(self)
        fvars = []
        for i in range(rel._signature.arity()):
            type = rel._signature._types[i]
            var = self.build_scalar_variable([type, tempvar(i)], {})
            fvars.append(var)

        self.head = LogicBuilder.build_atom(self, rel, fvars)
        self._populates = rel
        self._freevars = fvars
        self._annotations = []

    def annotate(self, a):
        self._annotations.append(a)
        return self

    # Elaborates the body of this rule by invoking a Python function that, when
    # supplied with the array of free variables declared by this rule will populate
    # the conjuncts that make up the body of the rule.
    #
    def elaborate(self, func):
        with self.genexists(): # type: ignore
            func(*(self.freevars()))
        return self

    def genexists(self):
        x = ExistentialConstraint.build_existential()
        self._existentials[id(x)] = x
        return x

    def freevars(self):
        return tuple(self._freevars)

    def populates(self): return self._populates

    def pprint(self):
        annotations = "\n".join([a.display() for a in self._annotations])
        head = self.head.pprint()
        body = self.rel_formula()
        rule = f"def {head}:\n{body}"
        if annotations != "":
            return annotations + "\n" + rule
        else:
            return rule

    # Returns a new Rule that duplicates self after mapping every
    # Relation referred to by an atom through relmap
    #
    def map(self, relmap):
        newrel = relmap[self.populates().entityid()]
        r = Rule(newrel)
        r._sconstraints = self._sconstraints
        r._atoms = {}
        for a in self._atoms:
            newatom = a.map(relmap)
            r._atoms[newatom.entityid()] = newatom
        r._existentials = {}
        for e in self._existentials.values():
            ext = e.map(relmap)
            r._existentials[ext.entityid()] = ext
        return r

def Vars(*args: Type):
    vars = []
    size = contextStack.size()
    for i, t in enumerate(args):
        if isinstance(t, RelVariable):
            vars.append(t)
        else:
            vars.append(t(f'_x{i}_{size}'))
    if len(vars) == 1:
        return vars[0]
    return tuple(vars)


from typing import List, Optional

from pandas import DataFrame
import relationalai as rai
from relationalai.early_access.dsl.ir.compiler import Compiler
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.rel.compiler import ModelToRel


class RelExecutor:

    @staticmethod
    def execute_model(model: Model, result_cols: Optional[List[str]] = None) -> DataFrame:
        ir_model = Compiler().compile_model(model)
        rel_model = str(ModelToRel().to_rel(ir_model))
        return RelExecutor.execute(model.name, rel_model, result_cols)

    @staticmethod
    def execute(database: str, rel_model: str, result_cols: Optional[List[str]] = None) -> DataFrame:
        """Executes Rel code using the RAI client."""

        resources = rai.clients.snowflake.Resources()
        resources.config.set("use_graph_index", False)

        try:
            resources.create_graph(database)
        except Exception as e:
            if "already exists" not in str(e).lower():
                raise e

        engine = resources.config.get("engine")
        results = resources.exec_raw(database, engine, rel_model)

        from relationalai.clients import result_helpers
        df, _ = result_helpers.format_results(results, None, result_cols)

        return df


from relationalai.early_access.dsl.core.exprs import _
from relationalai.early_access.dsl.core.relations import Relation, RelationSignature, rule
from relationalai.early_access.dsl.core.rules import Rule, Vars
from relationalai.early_access.dsl.core.types.standard import standard_value_types, Hash
from relationalai.early_access.dsl.types import Type


class AbstractEntityType(Type):

    def __init__(self, model, nm):
        super().__init__(model, nm)
        self._poprel = None

    def population(self):
        if self._poprel is None:
            self._poprel = Relation(self.namespace(), self.name(), RelationSignature(self))
        return self._poprel

    def ctor_name(self): return f"^{self.name()}"

    def entity(self): return True

    def addrule(self, rule):
        relation = self._model.external_relation(self.name(), self)
        if not isinstance(rule, Rule):
            # Assume that rule is a Python function that, when invoked can build a rule
            rule = Rule(relation).elaborate(rule)
        relation._rules.append(rule)
        return rule

    @staticmethod
    def root_unconstrained_type():
        return standard_value_types['Hash']

class EntityType(AbstractEntityType):

    def __init__(self, model, nm, *args):
        super().__init__(model, nm)
        self._domain = [*args]
        self._constructor = None

    def __xor__(self, args):
        return self.constructor()(*args)

    def build_relation_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in EntityType {self.name()}")

        if len(args) > 1:
            raise Exception(f"Cannot declare type variable of EntityType {self.name()} by supplying more than a type")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for EntityType {self.name()}")

        type = args[0]
        self._domain.append(type)
        return type

    def build_scalar_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in EntityType {self.name()}")

        if len(args) > 1:
            raise Exception(f"Cannot declare type variable of EntityType {self.name()} by supplying more than a type")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for EntityType {self.name()}")

        type = args[0]
        self._domain.append(type)
        return type

    # An EntityTypeConstructor is an n-ary functional Relation that maps tuples of
    # arity (n-1) to instances of the EntityType.
    #
    def constructor(self):
        if self._constructor is None:
            # Create a constructor function (relation) when this EntityType is created
            ctortype = self.ctordomain().copy()
            ctortype.append(self)
            self._constructor = Relation(self.namespace(),
                                         self.ctor_name(),
                                         RelationSignature(*tuple(ctortype)))
        return self._constructor

    def ctordomain(self):
        return self._domain

    def add_instance(self, val):
        if not isinstance(val, str):
            raise Exception(
                f"Attempt to add non-string instance {val} to entity type {self.name()}, but currently on string instances supported.")

        self.relation(f"is {val}", name=val)

        with self.__dict__[val]:
            @rule()
            def r(v): self^(val, v)

    def first(self, relation):
        with self:
            @rule()
            def r(e):
                relation(e, _)

    def filtered_by(self, filter, value):
        with self:
            @rule()
            def r(e):
                v = Vars(Hash)
                #
                value(v)
                filter(e, v)


    def qualified_name(self):
        return self._name

    def pprint(self):
        domain_str = ", ".join([f"x{i} in {t.name()}" for i, t in enumerate(self._domain)])
        return f"entity type {self.name()}({domain_str})"



from relationalai.early_access.dsl.state_mgmt import ComponentWrapper, StateManager, TransitionComponent, UndecoratedStateComponent, UndecoratedTransitionComponent

class Transition(StateManager):

    def __init__(self, sm):
        self._state_chart = sm

    # The transition that when fired activates this state machine from its
    # declared start state. Users override this method to declare more precise
    # effects of this operation.
    #
    def activate(self):
        return self.And(self.top(),
                        self._state_chart.inactive(),
                        self._state_chart.active().decorate())

    # An active transition starts and ends in the Active state. Most user-declared
    # transitions will be either activate or active transitions.
    #
    def active(self):
        return self.And(self.top(),
                        self.state_chart().active(),
                        self.state_chart().active().decorate())

    def after_state(self): return self._state_chart.after_state()
    def before_state(self): return self._state_chart.top()

    def top(self):
        return self.And(self.before_state(),
                        self.after_state())

    def state_chart(self): return self._state_chart

class SimpleTransition(Transition):

    def __init__(self, sm):
        super().__init__(sm)

    # The transition method of a SimpleTransition composes the transition
    # by conjoining before and after instances of the state machine being managed.
    # By default, these are the values returned by the before_state and after_state
    # methods respectively, but these may be overridden by keyword arguments 'before'
    # and 'after'.
    #
    def transition(self, **kwargs):
        if len(kwargs) > 2:
            raise Exception("Not possible")

        for a in kwargs:
            if a != 'before' and a != 'after':
                raise Exception("Only permissible keyword arguments to SimpleTransition.transition are 'before' and 'after'")
        if 'after' in kwargs:
            after_state = kwargs['after']
        else:
            after_state = self.after_state()

        if 'before' in kwargs:
            before_state = kwargs['before']
        else:
            before_state = self.before_state()

        return self.And(before_state, after_state)

class CompositeTransition(Transition):

    def __init__(self, sm):
        super().__init__(sm)
        self._components = {}

        comps = sm._components
        for c in comps:
            comp = comps[c]
            if isinstance(comp, UndecoratedStateComponent):
                self.add_undecorated_component(c, comp.machine().transition_manager())
            else:
                self.add_component(c, comp.machine().transition_manager())

    def activate(self):
        t = super().activate()

        transitions = []
        for c in self._components.values():
            transitions.append(c(c.manager().activate()))

        return self.And(t, *tuple(transitions))

    def add_component(self, compname, transm):
        state_mgr = self._state_chart
        state_components = state_mgr._components
        renamings = {}
        if compname in state_components:
            renamings = state_components[compname]._renamings

        self._components[compname] = TransitionComponent(compname, transm, renamings)
        self.__setattr__(compname, ComponentWrapper(transm, compname))

    def add_undecorated_component(self, compname, machine):
        state_mgr = self._state_chart
        state_components = state_mgr._components
        renamings = {}
        if compname in state_components:
            renamings = state_components[compname]._renamings
        self._components[compname] = UndecoratedTransitionComponent(compname, machine, renamings)
        self.__setattr__(compname, ComponentWrapper(machine, compname))

    def component(self, name): return self._components[name]

    def top(self):
        t = super().top()
        t2 = self.transition({})
        return self.And(t, t2)

    def transition(self, *transitions):
        component_transitions = {}
        for d in transitions:
           component_transitions = component_transitions | d

        transitions = []
        for c in self._components:
            comp = self._components[c]
            if c in component_transitions:
                transitions.append(comp(component_transitions[c]))
            else:
                transitions.append(comp(comp._prototype))
        return self.And(*tuple(transitions))




from typing import Optional

from relationalai.early_access.dsl.core.types import Type as CoreType
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.relations import Relation
from relationalai.early_access.dsl.core.utils import camel_to_snake
from relationalai.early_access.dsl.utils import build_relation_name


class Type(CoreType):

    # We can add relation components to a ConceptModule by invoking it
    # with arguments that interleave reading text with the Types used
    # to play various Roles
    #
    def __setattr__(self, key, value):
        if key in dir(self) and key not in self.__dict__:
            raise Exception(f"Cannot override method {key} of Type {self.name()} as an attribute.")
        else:
            if key[0] != '_':
                self._relations[key] = value
            return super().__setattr__(key, value)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __init__(self, model, nm):
        super().__init__(nm)
        self._model = model
        self._relations = {}
        self._generated_namespace = None

    def relation(self, *args, name: Optional[str] = None, namespace: Optional[Namespace]=None, functional: bool=False) -> Relation:
        self._generated_namespace = namespace
        relation = self._build_relation(name, args, functional)
        self.add_relation(relation)
        return relation

    def add_relation(self, relation: Relation):
        self._relations[relation._relname] = relation
        self.__setattr__(relation._relname, relation)

    def _build_relation(self, name, verbalization, functional):
        verb_parts = []
        roles:list[CoreType] = [self]
        if isinstance(verbalization, CoreType):
            roles.append(verbalization)
        else:
            for i in range(len(verbalization)):
                arg = verbalization[i]
                if isinstance(arg, str):
                    verb_parts.append(arg)
                else:
                    if isinstance(arg, CoreType):
                        roles.append(arg)
                    else:
                        raise Exception(
                            f"Predicate verbalization in Type {self.name()} is {arg}, which is neither a string nor a Type")

        if name is None:
            name = build_relation_name(roles, verb_parts)
        return self._model._create_relation(name, *roles, namespace=self.generated_namespace(), functional=functional).addverb(verb_parts)

    def generated_namespace(self):
        if self._generated_namespace is None:
            ename = camel_to_snake(self.name())
            self._generated_namespace = Namespace(ename, self.namespace())
        return self._generated_namespace


from typing import List, Optional

from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.ontologies.roles import AbstractRole
from relationalai.early_access.dsl.utils import build_relation_name, build_relationship_name


class Reading:

    # We initialize a Reading with a tuple that mixes text with Roles.
    #
    def __init__(self, *args, name: Optional[str] = None):
        self.rel_name = name  # Local name of Relation that materializes this Reading
        self.roles = []
        self.text_frags = []
        for i in range(len(args)):
            a = args[i]
            if isinstance(a, str):
                self.text_frags.append(a)
            else:
                if isinstance(a, AbstractRole):
                    self.roles.append(a)
                else:
                    raise Exception(f"Unknown Reading component {a} -- should be either text or a Role")

    def guid(self):
        return generate_stable_uuid(self.verbalize())

    def types(self) -> List[Type]:
        return [role.player() for role in self.roles]

    def template(self):
        temp_frags = []
        for i in range(len(self.roles)):
            temp_frags.append('{' + f"{i}" + '}')
            if i in range(len(self.text_frags)):
                temp_frags.append(self.text_frags[i])
        return " ".join(temp_frags)

    def verbalize(self):
        temp_frags = []
        for i in range(len(self.roles)):
            temp_frags.append(self.roles[i].verbalize())
            if i in range(len(self.text_frags)):
                temp_frags.append(self.text_frags[i])
        return " ".join(temp_frags)

    def to_rel_name(self):
        if self.rel_name:
            return self.rel_name
        return build_relation_name(self.roles, self.text_frags)

    def to_relationship_name(self):
        return build_relationship_name(self.roles, self.text_frags)




from collections import OrderedDict, defaultdict
from typing import Optional

from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.utils import generate_stable_uuid, camel_to_snake
from relationalai.early_access.dsl.ontologies.constraints import Unique, Mandatory
from relationalai.early_access.dsl.ontologies.readings import Reading
from relationalai.early_access.dsl.ontologies.roles import Role
from relationalai.early_access.dsl.types import Type as DslType
from relationalai.early_access.dsl.utils import extract_relation_text_with_signature


class Relationship:

    def __init__(self, model, *args, relation_name: Optional[str] = None):
        self._model = model
        self._is_subtype = False
        self._readings_map = OrderedDict()
        self._rolemap = OrderedDict()
        self._constraints = []
        if not contextStack.empty():
            tp = contextStack.root_context()
            if isinstance(tp, Namespace):
                self._namespace = tp
            else:
                self._namespace = Namespace.top
        else:
            self._namespace = Namespace.top
        if len(args) != 0:
            reading = self._reading_from_args(*args, relation_name=relation_name)
            for role in reading.roles:
                self._add_role(role)
            self._add_relation(reading)

    def __call__(self, *args, **kwargs):
        return self.relation(*args, **kwargs)

    def __enter__(self):
        contextStack.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()
        pass

    def __getitem__(self, key):
        return self._readings_map[key]

    def __setitem__(self, key, value):
        if key in self._readings_map:
            raise Exception(
                f'Cannot provide multiple Readings for the relation name {key} in Relationship {self._name}')
        else:
            if not isinstance(value, Reading):
                value = Reading(*value)

            if value.rel_name is not None:
                # [VAMI] TODO: check if we need this warning
                # warn(f'Overriding rel_name of Reading "{value.verbalize()}" from "{value.rel_name}" to "{key}"')
                pass
            self._readings_map[key] = value
            value.rel_name = key
            return value

    def __setattr__(self, key: str, value: Role) -> None:
        if key.startswith('_') or key in self.__dict__:
            super().__setattr__(key, value)
        elif isinstance(value, Role):
            value.part_of = self
            self._add_role(value)
            super().__setattr__(key, value)
        else:
            raise TypeError(f'Expected a Role instance for "{key}", got {type(value)}')

    def __getattr__(self, key):
        if key in self._rolemap:
            return self._rolemap[key]
        else:
            raise AttributeError(f'Relationship {self._name} has no Role named {key}')

    def _reading_from_args(self, *args, relation_name: Optional[str] = None):
        text_frags, sig = extract_relation_text_with_signature(*args)
        grouped_sig = defaultdict(list)
        for t in sig:
            grouped_sig[t.name()].append(t)
        reading_args = []
        index_map = {}
        # store indexes for roles player by the same Concept
        for t_name in grouped_sig:
            group_size = len(grouped_sig[t_name])
            if group_size > 1:
                index_map[t_name] = group_size
        for idx, t in enumerate(sig):
            reading_args.append(Role(t, self, idx, name=self._get_role_name(t, index_map)))
            if idx < len(text_frags):
                reading_args.append(text_frags[idx])
        return Reading(*reading_args, name=relation_name)

    def _get_role_name(self, t, index_map):
        name = None
        if t.name() in index_map:
            idx = index_map.get(t.name())
            index_map[t.name()] -= 1
            name = f"{camel_to_snake(t.name())}{idx}"
        return name

    def _name(self):
        return list(self._readings_map.values())[0].to_relationship_name()

    def _set_subtype(self):
        self._is_subtype = True
        return self

    def is_subtype(self):
        return self._is_subtype

    def guid(self):
        return generate_stable_uuid(self._name())

    def role(self, concept, name: Optional[str] = None, unique: bool = False, primary_key: bool = False,
             mandatory: bool = False) -> 'Relationship':
        role = Role(concept, pos=self.arity(), name=name, rel=self)
        self._add_role(role)
        setattr(self, role.ref_name(), role)

        # check constraints
        if unique and not primary_key:
            self.unique(role)
        if mandatory and not primary_key:
            self.mandatory(role)
        if primary_key:
            self.primary_key(role)
        return self

    def role_at(self, idx):
        if idx < 0 or idx >= self.arity():
            raise Exception(f'Role index {idx} out of bounds for Relationship {self._name}')
        return list(self._rolemap.values())[idx]

    def arity(self):
        return len(self._rolemap)

    def relation(self, *args, name: Optional[str] = None) -> 'Relationship':
        return self._add_relation(Reading(*args, name=name))

    def _add_role(self, rol):
        self._rolemap[rol.ref_name()] = rol
        return self

    def _add_reading(self, rdg):
        if not isinstance(rdg, Reading):
            raise Exception(f'Tried to add non-Reading {rdg} as a reading of Relationship {self._name}')

        self[rdg.to_rel_name()] = rdg

    def _add_relation(self, reading: Reading) -> 'Relationship':
        self._add_reading(reading)
        first_player = reading.roles[0].player()
        namespace = Namespace(camel_to_snake(first_player.name()), self._namespace)
        relation = self._model._create_relation(reading.to_rel_name(), *reading.types(), namespace=namespace).addverb(
            reading.text_frags)
        # do not add relations to standard types like String, Date, Integer etc.
        if isinstance(first_player, DslType):
            first_player.add_relation(relation)
        # sync relationship and roles with the model once relationship have at least 1 reading
        if len(self._readings_map) == 1:
            self._model._add_relationship(self)
            for c in self._constraints:
                self._model.constraint(c)
        return self

    def build_relation_variable(self, args, kwargs):
        return self.new_role(args, kwargs)

    def build_scalar_variable(self, args, kwargs):
        return self.new_role(args, kwargs)

    def new_role(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f'Unexpected error in Relationship {self._name}')

        if len(args) > 2:
            raise Exception(
                f'Cannot declare role of Relationship {self._name} by supplying more than a player type and a UUID')

        if len(kwargs) != 0:
            raise Exception(
                f'Cannot use keyword arguments when instantiating a type parameter for Relationship {self._name}')

        type = args[0]

        pre_bound_text = kwargs['pre_bound_text'] if 'pre_bound_text' in kwargs else None
        post_bound_text = kwargs['post_bound_text'] if 'post_bound_text' in kwargs else None

        if not isinstance(type, Type):
            raise Exception('Can only instantiate a Role with a Type as its player')
        role = Role(type, self, self.arity()).verbalization(pre_bound_text, post_bound_text)

        self._add_role(role)
        return role

    def readings(self):
        return list(self._readings_map.values())

    # Constraints

    def unique(self, *roles):
        """Add a uniqueness constraint on specified roles."""
        self._add_constraint(Unique(*roles))
        return self

    def mandatory(self, *roles):
        """Add a mandatory constraint on specified roles."""
        self._add_constraint(Mandatory(*roles))
        return self

    def primary_key(self, *roles):
        """Set the (composite) primary key & uniqueness constraint."""
        self._add_constraint(Unique(*roles, is_preferred_identifier=True))
        for r in roles:
            self._add_constraint(Mandatory(r))
        return self

    def _add_constraint(self, con):
        self._constraints.append(con)
        if not self.is_empty():
            self._model.constraint(con)

    # pretty printing readings
    def pprint(self):
        return f'relationship {self._name()}[{", ".join([r.verbalize() for r in self.readings()])}]'

    def is_empty(self):
        return len(self.readings()) == 0

class Attribute(Relationship):
    def __init__(
            self,
            model,
            concept,
            attr,
            mandatory: bool = False,
            primary_key: bool = False,
            reading_text: Optional[str] = None,
            reverse_reading_text: Optional[str] = None
    ):
        if reading_text is None:
            reading_text = 'has'

        super().__init__(model, concept, reading_text, attr)
        concept_role = self.role_at(0)
        self.unique(concept_role)
        if mandatory:
            self.mandatory(concept_role)

        attr_role = self.attr()
        if primary_key:
            self.primary_key(attr_role)

        if reverse_reading_text is not None:
            self.relation(attr_role, reverse_reading_text, concept_role)

    def attr(self):
        return self.role_at(1)

def _reading_to_relationship_name(reading: str) -> str:
    """
    Takes a readable `reading` string and returns the middle part of the Relationship name.

    This is done by removing spaces and converting the result to PascalCase."""
    return ''.join(word.capitalize() for word in reading.split(' '))



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Optional, Sequence as PySequence, cast
import itertools

import pandas as pd

from relationalai.clients import result_helpers
from relationalai.early_access.metamodel import factory as f, ir, builtins, types
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.builder.util import sanitize
from relationalai.early_access.rel.executor import RelExecutor

#--------------------------------------------------
# Global ID
#--------------------------------------------------

_global_id = itertools.count(0)

#--------------------------------------------------
# Producer
#--------------------------------------------------

class Producer():
    def __init__(self, model:Model) -> None:
        self._id = next(_global_id)
        self._model = model

    #--------------------------------------------------
    # Infix operator overloads
    #--------------------------------------------------

    def _bin_op(self, op, left, right) -> Expression:
        res = Var(self._model, None, self._model.Concept("Number"))
        return Expression(self._model, self._model._builtin_relationships[op], left, right, res)

    def __add__(self, other):
        return self._bin_op("+", self, other)
    def __radd__(self, other):
        return self._bin_op("+", other, self)

    def __mul__(self, other):
        return self._bin_op("*", self, other)
    def __rmul__(self, other):
        return self._bin_op("*", other, self)

    def __sub__(self, other):
        return self._bin_op("-", self, other)
    def __rsub__(self, other):
        return self._bin_op("-", other, self)

    def __truediv__(self, other):
        return self._bin_op("/", self, other)
    def __rtruediv__(self, other):
        return self._bin_op("/", other, self)

    def __floordiv__(self, other):
        return self._bin_op("//", self, other)
    def __rfloordiv__(self, other):
        return self._bin_op("//", other, self)

    def __pow__(self, other):
        return self._bin_op("^", self, other)
    def __rpow__(self, other):
        return self._bin_op("^", other, self)

    def __mod__(self, other):
        return self._bin_op("%", self, other)
    def __rmod__(self, other):
        return self._bin_op("%", other, self)

    def __neg__(self):
        return self._bin_op("*", self, -1)

    #--------------------------------------------------
    # Filter overloads
    #--------------------------------------------------

    def _filter(self, op, left, right) -> Expression:
        return Expression(self._model, self._model._builtin_relationships[op], left, right)

    def __gt__(self, other):
        return self._filter(">", self, other)
    def __ge__(self, other):
        return self._filter(">=", self, other)
    def __lt__(self, other):
        return self._filter("<", self, other)
    def __le__(self, other):
        return self._filter("<=", self, other)
    def __eq__(self, other) -> Any:
        return self._filter("=", self, other)
    def __ne__(self, other) -> Any:
        return self._filter("!=", self, other)

    #--------------------------------------------------
    # getattr
    #--------------------------------------------------

    def __getattr__(self, name:str) -> Any:
        if name.startswith("_"):
            raise AttributeError(f"{type(self).__name__} has no attribute {name}")
        if name not in self._relationships:
            any_concept = self._model.Concept("Any")
            self._relationships[name] = Relationship(
                (any_concept, any_concept), parent=self, prop_name=name, model=self._model
            )
        return self._relationships[name]

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, Relationship) and not name.startswith("_"):
            value._parent = self
            self._relationships[name] = value
        else:
            super().__setattr__(name, value)

    #--------------------------------------------------
    # Hash
    #--------------------------------------------------

    def __hash__(self) -> int:
        return hash(self._id)

    #--------------------------------------------------
    # Fallbacks
    #--------------------------------------------------

    def select(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.select` not implemented")

    def where(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.where` not implemented")

    def require(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.require` not implemented")

    def then(self, *args: Any):
        raise NotImplementedError(f"`{type(self).__name__}.then` not implemented")

    def alias(self, name: Optional[str] = None) -> Any:
        """
        Create an alias for this producer.

        If name is provided, creates a column alias for use in select statements.
        If name is not provided, creates a table alias (must be implemented by subclasses).
        """
        if name is not None:
            # Create an expression with the = operator and add an _alias_name attribute
            expr = Expression(self._model, self._model._builtin_relationships["="], self, Var(self._model, name))
            expr._alias_name = name  # Add this attribute to identify it as an alias expression
            return expr
        raise NotImplementedError(f"Table aliasing not implemented for {type(self).__name__}")

#--------------------------------------------------
# Concept
#--------------------------------------------------

class Concept(Producer):
    def __init__(self, name:str, model:Model):
        super().__init__(model)
        self._name = name
        self._alias_parent = self
        self._alias_count = 0
        self._alias_name = name
        self._relationships = {}

    def alias(self, name: Optional[str] = None) -> Any:
        if name is not None:
            return super().alias(name)

        c = Concept(self._name, self._model)
        parent = self._alias_parent
        c._alias_parent = parent
        parent._alias_count += 1
        c._alias_name = parent._name + str(parent._alias_count)
        return c

    def require(self, *args: Any) -> Fragment:
        return select(self).require(*args)

    # TODO maybe delete from_csv. Note from Sam:
    # Personally I'm in favor of a less coupled design. If we own the CSV ingestion, then we're going to be taking requests about adding certain kwargs, etc. If we let people do the ingestion on their own and just take the resulting data frame, then we separate concerns.
    def from_csv(self, path: str, keys: Any = None, rels: Any = None) -> Concept:
        return self.from_df(pd.read_csv(path), keys=keys, rels=rels)

    def from_df(self, df: pd.DataFrame, keys: Optional[list[str]] = None, rels: Optional[list[str]] = None) -> Concept:
        key_names = keys or df.columns
        rel_names = rels or [c for c in df.columns if c not in key_names]
        for row in df.itertuples(index=False):
            # construct entity using keys and derive properties
            self(
                **{k.lower(): getattr(row, k) for k in key_names}
            ).then(
                *[getattr(self, k.lower())(self, getattr(row, k)) for k in rel_names]
            )
        return self

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._model._construct(self, args, kwargs)

    def __str__(self):
        return self._alias_name

#--------------------------------------------------
# Relationship
#--------------------------------------------------

@dataclass
class Field():
    name:str
    type:Any

class Relationship(Producer):
    def __init__(self, *args:Any, parent:Producer|None, prop_name:str, model:Model):
        super().__init__(model)
        self._parent = parent
        self._name = prop_name
        self._prop_name = prop_name
        self._alias_parent = self
        self._alias_count = 0
        self._alias_name = prop_name
        self._relationships = {}
        self._fields:list[Field] = []
        for arg in args:
            if isinstance(arg, str) and not self._name:
                self._name = sanitize(arg).lower()
            else:
                self._fields.append(Field("", arg))
        model.relationships.append(self)

    def alias(self, name: Optional[str] = None) -> Any:
        if name is not None:
            return super().alias(name)

        r = Relationship(parent=self._parent, prop_name=self._prop_name, model=self._model)
        parent = self._alias_parent
        r._alias_parent = parent
        parent._alias_count += 1
        r._alias_name = parent._name + str(parent._alias_count)
        return r

    def __call__(self, *args: Any) -> Any:
        clean_args = list(args)
        if len(args) < len(self._fields):
            if self._parent:
                clean_args = [self._parent, *clean_args]
        return self._model._derive(self, clean_args)

    def __str__(self):
        if self._parent and self._prop_name:
            return f"{self._parent}.{self._alias_name}"
        return self._name

#--------------------------------------------------
# Expression
#--------------------------------------------------

class Expression(Producer):
    def __init__(self, model:Model, op:Relationship|Concept, *params:Any):
        super().__init__(model)
        self._op = op
        self._params = params

    def __str__(self):
        return f"({self._op} {' '.join(map(str, self._params))})"

    def __getattr__(self, name: str):
        raise AttributeError(f"Expression has no attribute {name}")

    # def where(self, *args: Any) -> Fragment:
    #     return where(*args, model=self._model)

    # def when(self, *args: Any) -> Fragment:
    #     return self._model.where(*args).then(self)

    def per(self, *args: Any) -> Expression:
        if isinstance(self._op, Aggregate):
            self._op._group = list(args)
            return self
        else:
            raise NotImplementedError(f"per is only supported for aggregates, not {type(self._op)}")

#--------------------------------------------------
# Var
#--------------------------------------------------

class Var(Producer):
    def __init__(self, model:Model, name:Optional[str]=None, type:Any=None):
        super().__init__(model)
        if not name:
            name = f"v{self._id}"
        self._name = name
        self._type = type

    def __str__(self):
        return str(self._name) if self._name else f"v{self._id}"

    def __getattr__(self, name: str):
        raise AttributeError(f"Var has no attribute {name}")

    def __hash__(self) -> int:
        return hash(self._id)

#--------------------------------------------------
# Fragment
#--------------------------------------------------

class Fragment():
    def __init__(self, parent:Fragment|None=None, kind:str="", items=None, model:Model|None=None):
        self._id = next(_global_id)
        self._stack = parent._stack + [self] if parent else [self]
        self._kind = kind
        self._items = list(items or [])
        self._model = parent._model if parent else model

    def select(self, *args: Any) -> Fragment:
        return select(*args, parent=self, model=self._model)

    def where(self, *args: Any) -> Fragment:
        return where(*args, parent=self, model=self._model)

    def require(self, *args: Any) -> Fragment:
        return require(*args, parent=self, model=self._model)

    def then(self, *args: Any) -> Fragment:
        return then(*args, parent=self, model=self._model)

    def _to_str(self):
        items = '\n  '.join(map(str, self._items))
        return f"{self._kind.upper()}\n  {items}"

    def __str__(self):
        return "\n".join([item._to_str() for item in self._stack])

    def __iter__(self):
        # Iterate over the rows of the fragment's results
        return self.to_df().itertuples(index=False)

    def inspect(self):
        # @TODO what format? maybe ignore row indices?
        print(self.to_df())

    def to_df(self):
        """Convert the fragment's results to a pandas DataFrame."""
        # @TODO currently this code assumes a Rel executor; should dispatch based on config
        if not self._model:
            raise ValueError("Cannot execute fragment without a model context")

        # from logscope import log
        # log(*self._model.rules)
        # log(self)

        builder_compiler = Compiler()
        rule_tasks = [builder_compiler.compile_fragment(rule, effects_only=True) for rule in self._model.rules]
        fragment_task = builder_compiler.compile_fragment(self)

        # log(*rule_tasks)
        # log(fragment_task)

        root = f.logical(rule_tasks + [fragment_task])
        ir_model = f.compute_model(root)

        executor = RelExecutor(self._model.name or "pyrel_main", dry_run=self._model._dry_run)
        results = executor.execute(ir_model)
        if isinstance(results, pd.DataFrame):
            return results
        else:
            # Convert results to DataFrame using the helper from snowflake.py
            df, _ = result_helpers.format_results(results, None)  # Pass None for task parameter
            return df


#--------------------------------------------------
# Select / Where
#--------------------------------------------------

def _get_model_or_error(args: tuple, model:Model|None=None) -> Model:
    if not model:
        # get model from first Producer or Fragment in args
        for arg in args:
            if isinstance(arg, (Producer, Fragment)) and arg._model:
                model = arg._model
                break
    if not model:
        raise ValueError("Cannot create fragment without a model context")
    return model

def select(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return _get_model_or_error(args, model=model).select(*args, parent=parent)

def where(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return _get_model_or_error(args, model=model).where(*args, parent=parent)

def require(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return _get_model_or_error(args, model=model).require(*args, parent=parent)

def then(*args: Any, parent:Fragment|None=None, model:Model|None=None) -> Fragment:
    return _get_model_or_error(args, model=model).then(*args, parent=parent)


#--------------------------------------------------
# Model
#--------------------------------------------------

class Model():
    def __init__(self, name:str, dry_run: bool = False):
        self._id = next(_global_id)
        self.name = name
        self._dry_run = dry_run
        self.concepts:dict[str, Concept] = {}
        self.relationships:list[Relationship] = []
        self._builtin_relationships:dict[str, Relationship] = {}
        self.rules:OrderedSet[Fragment] = ordered_set()
        self._import_builtins()

    def _import_builtins(self):
        for attribute_name in dir(builtins):
            builtin = getattr(builtins, attribute_name)
            if isinstance(builtin, ir.Relation):
                self._builtin_relationships[builtin.name] = \
                    Relationship(parent=None, prop_name=builtin.name, model=self)
        for attribute_name in dir(types):
            builtin = getattr(types, attribute_name)
            if isinstance(builtin, ir.ScalarType):
                # TODO - if this is a list/set/union, should we create a concept as well?
                self.Concept(builtin.name)

    def Concept(self, name:str) -> Concept:
        if name not in self.concepts:
            self.concepts[name] = Concept(name, model=self)
        return self.concepts[name]

    def Relationship(self, *args) -> Relationship:
        return Relationship(*args, parent=None, prop_name="", model=self)

    def Vars(self, *types) -> list[Var]:
        if len(types) == 1 and isinstance(types[0], int):
            return [Var(self) for _ in range(types[0])]
        elif len(types) == 1 and isinstance(types[0], str):
            items = types[0].split()
            name_type = [item.split(":") if ":" in item else [item, None] for item in items]
            return [Var(self, name=pair[0], type=pair[1]) for pair in name_type]
        return [Var(self, type=type) for type in types]

    def _add_fragment(self, kind:str, *args: Any, parent:Fragment|None=None) -> Fragment:
        fragment = Fragment(parent=parent, kind=kind, items=args, model=self)
        # Add rule to the model, but remove any incomplete sub-fragments from the rules first
        for item in fragment._stack + fragment._items:
            if isinstance(item, Fragment) and item in self.rules:
                self.rules.remove(item)
        self.rules.add(fragment)
        return fragment

    def where(self, *args: Any, parent=None) -> Fragment:
        """Create a where fragment in this model's context"""
        return self._add_fragment("where", *args, parent=parent)

    def select(self, *args: Any, parent=None) -> Fragment:
        """Create a select fragment in this model's context"""
        return self._add_fragment("select", *args, parent=parent)

    def require(self, *args: Any, parent=None) -> Fragment:
        """Create a require fragment in this model's context"""
        return self._add_fragment("require", *args, parent=parent)

    def then(self, *args: Any, parent=None) -> Fragment:
        """Create an effects fragment in this model's context"""
        return self._add_fragment("effects", *args, parent=parent)

    def _construct(self, con: Concept, args, kwargs: dict) -> Fragment:
        return self._add_fragment("construct", con, args, kwargs)

    def _derive(self, rel: Relationship, args) -> Fragment:
        return self._add_fragment("derive", rel, args)

    def count(self, *args):
        """Create a count aggregate"""
        return Aggregate(self, "count", *args)

    def sum(self, *args):
        """Create a sum aggregate"""
        return Aggregate(self, "sum", *args)

    def min(self, *args):
        """Create a min aggregate"""
        return Aggregate(self, "min", *args)

    def max(self, *args):
        """Create a max aggregate"""
        return Aggregate(self, "max", *args)

    def avg(self, *args):
        """Create an average aggregate"""
        return Aggregate(self, "avg", *args)


#--------------------------------------------------
# Compile
#--------------------------------------------------

class CompilerContext():
    def __init__(self):
        self.var_map:dict[Concept|Relationship|Expression|Var, ir.Var] = {}
        self.items:OrderedSet[ir.Task] = OrderedSet()
        self.group_contexts = {}

    def to_var(self, item:Concept|Relationship|Expression|Var) -> ir.Var:
        if item not in self.var_map:
            name = ""
            if isinstance(item, Var):
                name = item._name or ""
            elif isinstance(item, (Concept, Relationship)):
                name = item._alias_name or ""
                if isinstance(item, Relationship) and isinstance(item._parent, Concept):
                    name = f"{item._parent._alias_name}_{name}"
            name = name.lower()
            self.var_map[item] = f.var(name, types.Any)
        return self.var_map[item]

    def add(self, item:ir.Task):
        self.items.add(item)

    def clone(self):
        c = CompilerContext()
        c.var_map = self.var_map.copy()
        return c


class Compiler():
    def __init__(self):
        self.types:dict[Concept, ir.Type] = {}
        self.relations:dict[Relationship|Concept, ir.Relation] = {}

    def to_type(self, concept:Concept) -> ir.Type:
        if concept not in self.types:
            if concept._name in types.builtin_scalar_types_by_name:
                self.types[concept] = types.builtin_scalar_types_by_name[concept._name]
            else:
                self.types[concept] = f.scalar_type(concept._name)
        return self.types[concept]

    def to_relation(self, item:Concept|Relationship) -> ir.Relation:
        if item not in self.relations:
            if isinstance(item, Concept):
                fields = [f.field(item._name.lower(), self.to_type(item))]
                relation = f.relation(item._name, fields)
            elif isinstance(item, Relationship):
                if item._name in builtins.builtin_relations_by_name:
                    relation = builtins.builtin_relations_by_name[item._name]
                else:
                    fields = [f.field(cur.name, types.Any) for cur in item._fields]
                    relation = f.relation(item._name, fields)
            self.relations[item] = relation
            return relation
        else:
            return self.relations[item]

    def compile_effect(self, ctx:CompilerContext, effect) -> PySequence[ir.Value]:
        if isinstance(effect, Concept):
            res = f.derive(self.to_relation(effect), [ctx.to_var(effect)])
        elif isinstance(effect, Expression):
            args = []
            op = effect._op
            if isinstance(op, Concept):
                rel = self.to_relation(op)
                for param in effect._params:
                    subs = self.compile_action(ctx, param)
                    args.append(subs[-1])
                # kwargs = getattr(effect, '_kwargs', {})
                # for name, value in kwargs.items():
                #     args.append(f.lit(name))
                #     args.append(f.lit(value))
                res = f.derive(rel, args)
            elif isinstance(op, Relationship) and op._parent:
                parent = self.compile_action(ctx, op._parent)
                args.append(parent[-1])
                rel = self.to_relation(op)
                for param in effect._params:
                    subs = self.compile_action(ctx, param)
                    args.append(subs[-1])
                res = f.derive(rel, args)
            else:
                rel = self.to_relation(op)
                for param in effect._params:
                    subs = self.compile_action(ctx, param)
                    args.append(subs[-1])
                res = f.derive(rel, args)
        elif isinstance(effect, Fragment):
            if effect._kind == "effects" and len(effect._items) == 1:
                return self.compile_effect(ctx, effect._items[0])
            elif effect._kind == "construct":
                return [self.compile_construct(ctx, *effect._items)]
            elif effect._kind == "derive":
                return [self.compile_derive(ctx, *effect._items)]
            else:
                raise ValueError(f"Cannot compile effect with fragment: {effect}")
        else:
            raise ValueError(f"Cannot compile effect: {effect}")
        ctx.add(res)
        return res.args

    def compile_construct(self, ctx: CompilerContext, con: Concept, args, kwargs: dict) -> ir.Var:
        # TODO handle args
        typ = cast(ir.ScalarType, self.to_type(con))
        var = ctx.to_var(con)
        ctx.add(f.derive(f.entity(typ), [var]))
        props = {}
        for k, v in kwargs.items():
            prop = f.property(k, con._name, typ, k, types.Any)
            comp_v = self.compile_action(ctx, v)[-1]
            props[prop] = comp_v
            ctx.add(f.derive(prop, [var, comp_v]))
        ctx.add(f.construct(var, props))
        return var

    def compile_derive(self, ctx: CompilerContext, rel: Relationship, args) -> ir.Var:
        assert len(args) <= 2 # TODO
        if len(args) == 1:
            key = rel._parent
            assert key and isinstance(key, Concept)
            val = args[0]
        else:
            key = args[0]
            val = args[1]
        prop = f.property(rel._name, key._name, self.to_type(key), rel._name, types.Any)
        val = self.compile_action(ctx, val)[-1]
        var = ctx.to_var(key)
        ctx.add(f.derive(prop, [var, val]))
        return var

    def compile_aggregate(self, ctx: CompilerContext, agg: Aggregate) -> ir.Var:
        """Compile an aggregate expression into IR"""
        agg_relation = f.relation(agg._name, [
            f.field("result", types.Any)
        ])

        # Handle projection vars and arg to be aggregated over
        proj_vars = []
        new_args = []
        if agg._name == "count" and not agg._args:
            new_args.append(f.lit(1))
        else:
            for i, arg in enumerate(agg._args):
                subs = self.compile_action(ctx, arg)
                if subs:  # Make sure subs is not None and not empty
                    if i < len(agg._args) - 1:
                        var = subs[-1]
                        if not isinstance(var, ir.Var):
                            # Only vars are supported in the projection list
                            var = f.var("proj", types.Any)
                            ctx.add(f.lookup(builtins.eq, [var, subs[-1]]))
                        proj_vars.append(subs[-1])
                    else:
                        # The last arg is the one to be aggregated over, and goes in the args list
                        new_args.append(subs[-1])
                else:
                    raise ValueError(f"Failed to compile aggregate argument: {arg}")

        # Handle group-by vars
        group_vars = []
        for group_var in agg._group:
            subs = self.compile_action(ctx, group_var)
            if subs:  # Make sure subs is not None and not empty
                var = subs[-1]
                if not isinstance(var, ir.Var):
                    # Only vars are supported in the group list
                    var = f.var("group", types.Any)
                    ctx.add(f.lookup(builtins.eq, [var, subs[-1]]))
                group_vars.append(var)
            else:
                raise ValueError(f"Failed to compile group variable: {group_var}")

        result_var_name = f"v{agg._id}"
        result_var = f.var(result_var_name, types.Any)
        new_args.append(result_var)

        agg_task = f.aggregate(agg_relation, proj_vars, group_vars, new_args)
        ctx.add(agg_task)

        if agg._where:
            for action in agg._where._items:
                self.compile_action(ctx, action)

        return result_var

    def compile_action(self, ctx: CompilerContext, action) -> PySequence[ir.Value]:
        if isinstance(action, Concept):
            res = f.lookup(self.to_relation(action), [ctx.to_var(action)])
        elif isinstance(action, Expression):
            rel = self.to_relation(action._op)
            args = []
            for param in action._params:
                subs = self.compile_action(ctx, param)
                args.append(subs[-1])
            if len(args) < len(rel.fields):
                args.append(f.var("", types.Any))
            res = f.lookup(rel, args)
        elif isinstance(action, Var):
            return [ctx.to_var(action)]
        elif isinstance(action, (str, int, float)):
            return [f.lit(action)]
        elif isinstance(action, Relationship):
            if isinstance(action, Aggregate):
                result_var = self.compile_aggregate(ctx, action)
                return [result_var]
            elif action._parent:
                parent = self.compile_action(ctx, action._parent)
                res = f.lookup(self.to_relation(action), [parent[-1], ctx.to_var(action)])
            else:
                raise ValueError(f"Cannot compile parentless relationship: {action}")
        elif isinstance(action, Fragment):
            if action._kind == "effects" and len(action._items) == 1:
                return self.compile_action(ctx, action._items[0])
            else:
                raise ValueError(f"Cannot compile action with fragment: {action}")
        elif isinstance(action, GroupingContext):
            return []
        else:
            raise ValueError(f"Cannot compile action {action} of type {type(action.__class__.__name__)}")

        if res is None:
            raise ValueError(f"Failed to compile action: {action}")

        ctx.add(res)
        return res.args

    def compile_fragment(self, fragment:Fragment, effects_only:bool=False) -> ir.Task:
        ctx = CompilerContext()
        reordered = []
        cur_effect = None

        ctx.group_contexts = {}

        for item in fragment._stack:
            if item._kind == "where":
                for action in item._items:
                    if isinstance(action, GroupingContext):
                        ctx.group_contexts[id(action)] = action
                    elif isinstance(action, Var) and hasattr(action, "_value") and isinstance(action._value, GroupingContext):
                        ctx.group_contexts[action._name] = action._value

        for item in fragment._stack:
            if item._kind in ["select", "require", "effects"]:
                cur_effect = item
            else:
                reordered.append(item)
        if cur_effect:
            reordered.append(cur_effect)

        for item in reordered:
            if item._kind == "select":
                aliases = []
                for ix, action in enumerate(item._items):
                    # Check whether this is an alias expression
                    if isinstance(action, Expression) and hasattr(action, '_alias_name'):
                        value_expr = action._params[0]
                        value_res = self.compile_action(ctx, value_expr)
                        if value_res and len(value_res) > 0:
                            aliases.append((action._alias_name, value_res[-1]))
                        else:
                            raise ValueError(f"Failed to compile alias expression: {action}")
                    else:
                        # Regular expression
                        res = self.compile_action(ctx, action)
                        if res and len(res) > 0:
                            name = ix
                            if isinstance(res[-1], ir.Var):
                                name = res[-1].name or f"v{res[-1].id}"
                            aliases.append((name, res[-1]))
                        else:
                            raise ValueError(f"Failed to compile select item: {action}")
                ctx.add(f.output(aliases))
            elif item._kind == "where":
                for action in item._items:
                    # Skip GroupingContext objects in where clauses
                    if isinstance(action, GroupingContext):
                        continue
                    try:
                        self.compile_action(ctx, action)
                    except Exception as e:
                        raise ValueError(f"Error compiling where clause action {action}: {e}")
            elif item._kind == "require":
                pass
            elif item._kind == "effects":
                for action in item._items:
                    self.compile_effect(ctx, action)
            elif item._kind == "construct":
                self.compile_construct(ctx, *item._items)
            elif item._kind == "derive":
                self.compile_derive(ctx, *item._items)

        if effects_only and not any(isinstance(i, ir.Update) for i in ctx.items.list):
            # If effects_only is true, only tasks involving updates
            # TODO this may not be a sufficient check - what if there are updates nested deeper inside?
            return f.logical([]) # TODO or f.success() but this causes error currently

        return f.logical(list(ctx.items))

#--------------------------------------------------
# Aggregate
#--------------------------------------------------

class Aggregate(Relationship):
    def __init__(self, model: Model, name: str, *args: Any):
        super().__init__(parent=None, prop_name=name, model=model)
        self._name = name
        self._args = args
        self._projection = []
        self._group = []
        self._where = None

    def per(self, *args: Any) -> Aggregate:
        """Specify the grouping variables for this aggregate"""
        self._group = list(args)
        return self

    def where(self, *args: Any) -> Aggregate:
        """Add filters to the aggregation"""
        # Create a where fragment that will be applied to this aggregate
        fragment = Fragment(parent=None, kind="where", items=args, model=self._model)
        self._where = fragment
        return self

    def alias(self, name: Optional[str] = None) -> Any:
        """Create an alias for this aggregate"""
        if name is not None:
            # Create an expression with the = operator and add an _alias_name attribute
            expr = Expression(self._model, self._model._builtin_relationships["="], self, Var(self._model, name))
            expr._alias_name = name  # Add this attribute to identify it as an alias expression
            return expr
        return super().alias(name)

    def __str__(self):
        args_str = ", ".join(map(str, self._args))
        group_str = ", ".join(map(str, self._group)) if self._group else ""
        return f"{self._name}({args_str}){' per ' + group_str if group_str else ''}"

def _create_aggregate_function(name: str):
    """Create an aggregate function with the given name"""
    def aggregate_function(*args: Any, model: Optional['Model'] = None) -> Aggregate:
        model = _get_model_or_error(args, model=model)
        return Aggregate(model, name, *args)
    return aggregate_function

# Define the standard aggregate functions
count = _create_aggregate_function("count")
sum = _create_aggregate_function("sum")
min = _create_aggregate_function("min")
max = _create_aggregate_function("max")
avg = _create_aggregate_function("avg")

class GroupingContext:
    def __init__(self, model: 'Model', *group_keys: Any):
        self._model = model
        self._group_keys = group_keys
        # Add a unique ID to help with identification
        self._id = next(_global_id)

    def __str__(self):
        keys_str = ", ".join(str(key) for key in self._group_keys)
        return f"per({keys_str})"

    def count(self, *args: Any) -> Aggregate:
        if not args:
            if self._group_keys:
                agg = Aggregate(self._model, "count", self._group_keys[0])
            else:
                agg = Aggregate(self._model, "count", 1)
        else:
            agg = Aggregate(self._model, "count", *args)

        agg._group = list(self._group_keys)
        return agg

    def sum(self, *args: Any) -> Aggregate:
        agg = Aggregate(self._model, "sum", *args)
        agg._group = list(self._group_keys)
        return agg

    def avg(self, *args: Any) -> Aggregate:
        agg = Aggregate(self._model, "avg", *args)
        agg._group = list(self._group_keys)
        return agg

    def min(self, *args: Any) -> Aggregate:
        agg = Aggregate(self._model, "min", *args)
        agg._group = list(self._group_keys)
        return agg

    def max(self, *args: Any) -> Aggregate:
        agg = Aggregate(self._model, "max", *args)
        agg._group = list(self._group_keys)
        return agg

def per(*args: Any, model: Optional['Model'] = None) -> GroupingContext:
    model = _get_model_or_error(args, model=model)
    return GroupingContext(model, *args)

def distinct(value: Any) -> 'Expression':
    """Create a distinct expression that removes duplicates"""
    if isinstance(value, Producer):
        model = value._model
        return Expression(model, model._builtin_relationships["distinct"], value)
    else:
        raise ValueError(f"Cannot create distinct expression from {value}")

__all__ = ["select", "where", "require", "then", "distinct", "per", "count", "sum", "min", "max", "avg"]

#--------------------------------------------------
# Todo
#--------------------------------------------------
"""
- [] construct
- [] static data handling
- [] Quantifiers
    - [] not
    - [] exists
    - [] forall
- [] Aggregates
- [] Require
- [] Multi-step chaining
- [] oneof
- [] anyof
- [] query when iterating over a select
- [] capture all rules
- [] implement aliasing
    - [] Concept
    - [] Relationship
    - [] Expression
- [] support defining relationships via madlibs Relationship("{Person} was born on {birthday:Date}")
- [] handle relationships with multiple name fields being accessed via prop:
    Package.shipment = Relationship("{Package} in {Shipment} on {Date}")
    Package.shipment.date, Package.shipment.shipment, Package.shipment.package
- [] import handling
    - [] table
    - [] csv

"""



"""
    Elementary IR types.
"""
from . import ir, util
import sys
from typing import cast


#
# Basic Types
#
Null = ir.ScalarType("Null")
Any = ir.ScalarType("Any")
Hash = ir.ScalarType("Hash")
String = ir.ScalarType("String")
Number = ir.ScalarType("Number")
Int = ir.ScalarType("Int")
Decimal = ir.ScalarType("Decimal")
Bool = ir.ScalarType("Bool")
Binary = ir.ScalarType("Binary") # 0 or 1
Symbol = ir.ScalarType("Symbol")

AnySet = ir.SetType(Any)
NumberSet = ir.SetType(Number)
StringSet = ir.SetType(String)
IntSet = ir.SetType(Int)

AnyList = ir.ListType(Any)
SymbolList = ir.ListType(Symbol)

def is_builtin(t: ir.Type):
    return t in builtin_types

def _compute_builtin_types() -> list[ir.Type]:
    module = sys.modules[__name__]
    types = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Type):
            types.append(builtin)
    return types

builtin_types = _compute_builtin_types()
builtin_scalar_types_by_name = dict((t.name, t) for t in cast(list[ir.ScalarType], util.filter_by_type(builtin_types, ir.ScalarType)))



from abc import abstractmethod

from relationalai.early_access.dsl.core.builders.logic import ScalarConstraintBuilder, LogicBuilder
from relationalai.early_access.dsl.core.constraints.scalar import reduce_constraints, unsatisfiable
from relationalai.early_access.dsl.core.cset import ConstraintSet
from relationalai.early_access.dsl.core.exprs.scalar import ScalarExprBuilder, box_number
from relationalai.early_access.dsl.core.relations import Relation, RelationSignature
from relationalai.early_access.dsl.core.types import AbstractValueType
from relationalai.early_access.dsl.core.types.variables import UnstableConstraintSet, NominalTypeVariable, \
    BasicTypeVariable, TypeVariable


# A ConstrainedValueType:
#
#  * is identified by a *name*,
#  * is parameterized by a sequence *params* of TypeVariables
#  * has a set *constraints* of ScalarConstraints
#
# We implement constraint sets using an int-keyed dictionary to simplify
# duplicate constraint removal.
#
class ConstrainedValueType(AbstractValueType, ScalarExprBuilder, LogicBuilder):

    def __enter__(self): # type: ignore
        super().__enter__()
        return self.typevar_tuple()

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        self.name_type_variables()
        constraints = self.denotes()
        if unsatisfiable(constraints):
            raise Exception(f"Value type {self.name()} has an unsatisfiable set of type constraints")

    def __init__(self, nm):
        AbstractValueType.__init__(self, nm)
        self._params = []
        self._type_constraints = UnstableConstraintSet()
        self._constructor = None

    def __xor__(self, args):
        return self.constructor()(*args)

    def build_comparison(self, left, op, right):
        c = ScalarConstraintBuilder.build_comparison(self, left, op, right)
        tc = self._type_constraints
        tc.add_scalar_constraint(c)
        return c

    # Generates a string that displays the TypeVariable constraints (one per line)
    # declared by this ConstrainedValueType
    #
    def constraint_display(self):
        exists = []
        for v in self._params:
            if isinstance(v, NominalTypeVariable):
                exists.append(v.all_constraint_display())
        exists_atoms = "\n".join(exists)
        exists_ppr = exists_atoms + "\n" if exists_atoms != "" else ""

        constraints = [f"   {c.pprint()}" for c in self.verifies().values()]
        if exists_ppr != "":
            return exists_ppr + "\n" + "\n".join(constraints)
        else:
            return "\n".join(constraints)

    def constructor(self):
        if self._constructor is None:
            # Create a constructor function (relation) for this ValueType
            ctortype = [item.typeof() for item in self.ctordomain()]
            ctortype.append(self)
            self._constructor = Relation(self.namespace(),
                                         self.ctor_name(),
                                         RelationSignature(*tuple(ctortype)))
        return self._constructor

    # Returns True if the Python constant `val` is an instance of this ValueType
    def contains(self, val):

        if self.signature_arity() != 1:
            raise Exception(
                f"NOT IMPLEMENTED: Check containment of values in ({self.name()}) which is not unary")

        # Get the lone type variable over which `self` imposes additional constraints
        tvar = self._params[0]
        if tvar.decl.contains(val) and isinstance(self._type_constraints, ConstraintSet):
            return self._type_constraints.evaluate({tvar.varname(): box_number(val)}) is not None

        return False

    def ctordomain(self):
        return self._params

    def ctor_name(self):
        return f"^{self.name()}"

    # Returns what ORM would consider to be the "DataType" of this AbstractValueType, if it makes sense.
    #
    def datatype(self):
        if not self.unary():
            raise Exception(
                f"Cannot infer an ORM datatype from AbstractValueType {self.name()}, which is not a "
                f"ConstrainedUnaryValueType")

        return self._params[0].typeof()

    # Assemble the set of constraints that are either declared or inherited by
    # this ConstrainedValueType. This method does not attempt to minimize constraints,
    # in contrast with verifies.
    #
    def denotes(self):
        C = self._type_constraints._sconstraints.copy()
        for ac in self._type_constraints._atoms.values():
            scs = self.interpret_type_atom(ac)
            for s in scs:
                if s not in C:
                    C[s] = scs[s]

        for p in self._params:
            C = p.denotes(C)

        return C

    # Generates the headline declaration of this ConstrainedValueType, separately
    # form the constraints, as the structure of this headline varies based on
    # whether this is a ValueType or a ValueSubtype
    #
    @abstractmethod
    def head_display(self) -> str: pass

    # Interpret Atom ac as an AtomicConstraint of the form:
    #
    #    T(x)
    #
    # where T is a ConstrainedValueType and x is a TypeVariable.
    #
    # Checks that the type of x conforms to that of the lone TypeVariable
    # that parameterizes T and returns all ScalarConstraints that T imposes
    # on those variables.
    #
    def interpret_type_atom(self, ac):

        T = ac.relation()
        args = ac.args

        if len(args) > 1:
            raise Exception(f"Cannot apply ValueSubtype constraint ({T.name()}) to more than one TypeVariable")

        vmap = {}
        tv = args[0]
        if isinstance(tv, BasicTypeVariable):
            if T.nominal():
                raise Exception(
                        f"Argument mismatch when using type {T.name()} as an AtomicConstraint in context of ValueType {self.name()}")

            if T.root_unconstrained_type() != tv.decl.root_unconstrained_type():
                raise Exception(
                    f"Applying type constraint {T.name()} to variable of incompatible unconstrained "
                    f"type {tv.decl.physical_typeof().pprint()}")

            vmap[id(T._params[0])] = tv
        else:
            # tv is a NominalTypeVariable

            ttup = flatten_typevar_tuple([], T.typevar_tuple())
            tvtup = flatten_typevar_tuple([], tv.typevar_tuple())

            if len(ttup) != len(tvtup):
                raise Exception(
                    f"Argument mismatch when using type {T.name()} as an AtomicConstraint in context of ValueType {self.name()}")

            for i in range(len(ttup)):
                if not isinstance(tvtup[i], TypeVariable):
                    raise Exception(
                        f"When using type {T.name()} as an AtomicConstraint in context of ValueType {self.name()}, {i}-th "
                        f"argument is not a TypeVariable")

                tup_basic_type = ttup[i].physical_typeof()
                tvtup_basic_type = tvtup[i].physical_typeof()

                # Validate that tup_basic_type and arg_basic_type are compatible

                if tup_basic_type != tvtup_basic_type:
                    raise Exception(
                        f"Applying basic type constraint {T.name()} to variable of incompatible unconstrained "
                        f"type {tvtup_basic_type.pprint()}")

                vmap[id(ttup[i])] = tvtup[i]

        rv = {}
        constraints = T.verifies()
        for cindex in constraints:
            c = constraints[cindex]
            dupc = c.revar(vmap)
            rv[dupc.entityid()] = dupc
        return rv

    # Assign a unique number to each type variable (including the parts of nominal type variables,
    # recursively) so that each type variable in the signature of this ConstrainedValueType will
    # have a unique name to use in interpreting type constraints.
    #
    def name_type_variables(self, initial_pos=0):
        pos = initial_pos
        for v in self._params:
            pos = v.setpos(pos)
        self._type_constraints = self._type_constraints.stabilize()

    def pprint(self):
        body = self.constraint_display()
        if body != "":
            return self.head_display() + ":\n" + body
        else:
            return self.head_display()

    def signature_arity(self):
        sum = 0
        for v in self._params:
            sum = sum + v.typeof().signature_arity()
        return sum

    # Returns the (possibly nested) tuple of BasicTypeVariables over which this ConstrainedValueType
    # ranges. This is useful when instantiating this type, as the calling context will often need to
    # directly reference these BasicTypeVariables. So, for instance, using this facility, we can
    # write code like:
    #
    #    ((x, y), (w, z)) = DoubleInterval.typevar_tuple()
    #
    # The flatten_typevar_tuple function can be used to flatten these tuples into an array, in this
    # case:
    #
    #    [ x, y, w, z ]
    #
    def typevar_tuple(self):
        tvars = []
        for tvar in self._params:
            tp = tvar.typeof()
            if tp.nominal():
                tvars.append(tvar.typevar_tuple())
            else:
                tvars.append(tvar)
        if len(tvars) == 1:
            return tvars[0]
        else:
            return tuple(tvars)

    # Helper method that returns the UnconstrainedSignature of this ConstrainedValueType
    # in the special case where that tuple is a single BasicTypeVariable, raising
    # an exception otherwise.
    #
    def value(self):
        if self.signature_arity() != 1:
            raise Exception(
                f"Cannot refer to singleton value type variable for non-unary AbstractValueType {self.name()}")
        else:
            return self.typevar_tuple()

    # Returns that subset of those ScalarConstraints that this ConstrainedValueType
    # denotes that we actually need to verify to ensure that values in its signature
    # satisfy the constraints.
    #
    def verifies(self):
        return reduce_constraints(self.denotes())

    def constraint(self) -> bool:
        return True


def flatten_typevar_tuple(seq, tup):
    if type(tup) is tuple:
        for v in tup:
            if tuple(v):
                seq = flatten_typevar_tuple(seq, v)
            else:
                seq.append(v)
    else:
        seq.append(tup)

    return seq



from relationalai.early_access.dsl.core import warn
from relationalai.early_access.dsl.core.constraints import Constraint, TrueConstraint, FalseConstraint, diffof

from z3 import Bool, Int, Real, Solver, unsat

class ScalarConstraint(Constraint):

    def __init__(self, x, y, z):
        self.op = y
        if y == "=" and x.grounded():
            self.left = z
            self.right = x
        else:
            self.left = x
            self.right = z

    # A binding is a definition in which the right-hand side of the equality
    # comparison is grounded and so will evaluate to a constant (e.g., "x == 20")
    #
    def binding(self):
        return self.definition() and self.right.grounded()

    # This ScalarConstraint is checkable in static verification contexts because
    # it can be trivially translated into a Z3 constraint
    #
    def checkable(self):
        return self.left.checkable() and self.right.checkable()

    def commutative_definition(self):
        return self.definition() and self.right.variable()

    def commute(self):
        if not self.commutative_definition():
            raise Exception(f"Cannot commute non-commutative definition {self.pprint()}")
        return ScalarConstraint( self.right, '=', self.left )

    # A definition is a left var ScalarConstraint of the form that involves an
    # equality operator (e.g., "x == 20", or "x == y + 5" )
    #
    def definition(self):
        return self.left.variable() and self.equality()

    def equality(self):
        return self.op == "="

    def negate(self):
        left = self.left
        right = self.right
        if self.op == "=":
            return ScalarConstraint(left, "!=", right)
        if self.op == "!=":
            return ScalarConstraint(left, "=", right)
        if self.op == "<=":
            return ScalarConstraint(left, ">", right)
        if self.op == "<":
            return ScalarConstraint(left, ">=", right)
        if self.op == ">=":
            return ScalarConstraint(left, "<", right)
        if self.op == ">":
            return ScalarConstraint(left, "<=", right)

        raise Exception(f"Cannot negate constraint {self.pprint()}")

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):

        left = self.left
        right = self.right
        result = []

        if hasattr(left, "display"):
            result.append(left.display())
        else:
            result.append(left)

        if self.op == "=":
            result.append("==")
        else:
            result.append(self.op)

        if hasattr(right, "display"):
            result.append(right.display())
        else:
            result.append(right)

        return " ".join(result)

    def rel_formula(self):
        return f"{self.left.display()} {self.op} {self.right.display()}"

    def grounded(self):
        return self.left.grounded() and self.right.grounded()

    def grounds(self, v, groundings):
        if self.equality():
            lv = self.left
            if lv.variable():
                if lv.entityid() == v.entityid():
                    return self.right.grounded_using(groundings)
                else:
                    rv = self.right
                    if rv.variable() and rv.entityid() == v.entityid():
                        return self.left.grounded_using(groundings)
        return False

    def scalar_refs(self):
        dic = self.left.scalar_refs()
        rdict = self.right.scalar_refs()
        for v in rdict:
            if v not in dic:
                dic[v] = rdict[v]
        return dic

    def refersto(self, varname: str) -> bool:
        return self.left.refersto(varname) or self.right.refersto(varname)

    def relational(self) -> bool:
        return self.left.relational() or self.right.relational()

    def rename(self, renaming):
        return ScalarConstraint(self.left.rename(renaming),
                                self.op,
                                self.right.rename(renaming))

    def revar(self, vmap):
        return ScalarConstraint(self.left.revar(vmap),
                                self.op,
                                self.right.revar(vmap))

    # Report out the propositional variables that are used in this constraint as
    # an extension to the bindings dictionary that is provided.
    #
    def report_propositional_variable_usage(self, bindings):
        uses = self.scalar_refs()

        if len(bindings) == 0:
            return uses
        else:
            bdgs = bindings.copy()
            for v in uses:
                if v not in bindings:
                    bdgs[v] = uses[v]

            return bdgs

    def simplify(self):
        if self.grounded():
            exec(f"self.x = {self.pprint()}")
            if self.x: # type: ignore
                return TrueConstraint()
            else:
                return FalseConstraint()
        else:
            left = self.left.simplify()
            right = self.right.simplify()
            if left.grounded() and self.equality():
                return ScalarConstraint(right, self.op, left)
            else:
                return ScalarConstraint(left, self.op, right)

    def substitute(self, bindings):
        x = self.left.substitute(bindings)
        y = self.right.substitute(bindings)
        return ScalarConstraint(x, self.op, y)

    def to_z3(self, vars):
        left = self.left.to_z3(vars)
        right = self.right.to_z3(vars)
        op = self.op

        if op == "=":
            return left.__eq__(right)
        if op == "!=":
            return left.__ne__(right)
        if op == ">=":
            return left.__ge__(right)
        if op == ">":
            return left.__gt__(right)
        if op == "<=":
            return left.__le__(right)
        if op == "<":
            return left.__lt__(right)

        raise Exception(f"Cannot convert {self.pprint()} into a z3 constraint")

    def entityid(self):
        return hash((self.op,
                     self.left.entityid(),
                     self.right.entityid()))


# Given a dictionary 'props' of ScalarConstraints, return the dictionary
# containing that subset of props that are binding definitions
#
def binding_definitions_of(props):
    return {c.entityid(): c for c in props.values() if c.binding()}


# Given a set props of ScalarConstraints, partition props into
# (bcons, other) where bcons are binding definitions, and return
# (bdgs, bcons, other) where bdgs is a binding dictionary used to
# apply constraints from bcons.
#
def bindings_from(props):
    bdgs = {}
    bcons = {}
    other = []

    bprops = binding_definitions_of(props)

    # Partition props into a dictionary of substitution bindings (bdgs)
    # that derive from the binding definitions in props and any other
    # (non-binding) constraints in props.
    #
    for chash in props:
        c = props[chash]
        if chash in bprops:
            v = c.left.display()
            if v in bdgs:
                # Then we have (at least 2) constraints of the form:
                #    v == expr1
                #    v == expr2
                # for expressions expr1 and expr2. In this case, we
                # need to add the equality constraint `expr1 == expr2`
                # to other.
                #
                expr1 = bdgs[v]
                expr2 = c.right
                if expr2.variable():
                    # Prefer to put the variable on the LHS when possible
                    #
                    other.append(ScalarConstraint(expr2, "=", expr1))
                else:
                    other.append(ScalarConstraint(expr1, "=", expr2))
            else:
                bdgs[v] = c.right
                bcons[chash] = c
        else:
            other.append(c)

    return bdgs, bcons, other

def commutative_definitions_from(props):
    defns = {}
    cdcons = {}
    other = []

    cdefs = commutative_definitions_of(props)

    # Partition props into a dictionary of commutative definitions (defns)
    # that derive from the binding definitions in props and any other
    # (non-binding) constraints in props.
    #
    for chash in props:
        c = props[chash]
        if chash in cdefs and c.commute().entityid() not in defns:
            # Add {chash: c} to defns if its commuted equivalent not already there!
            defns[c.left.display()] = c.right
            cdcons[chash] = c
        else:
            other.append(c)

    return defns, cdcons, other


# Given a set C of ScalarConstraints, return that subset that are checkable
#
def checkable(C):
    tocheck = {}
    for c in C:
        constraint = C[c]
        if constraint.checkable():
            tocheck[c] = constraint
        else:
            warn("Attempt to check constraint set involving an uncheckable scalar constraint. Expected in rules with aggregations")
    return tocheck


def definitions_of(props):
    return {c.entityid(): c for c in props.values() if c.definition()}

def commutative_definitions_of(props):
    return {c.entityid(): c for c in props.values() if c.commutative_definition()}

# Given a set C of ScalarConstraints and a singleton ScalarConstraint c
# where c not in C, returns true if C entails c, i.e., if:
#
#     [[ C ]] \subseteq [[ C u {c} ]]
#
def entails(C, c):

    if not c.checkable():
        warn(f"Attempt to check entailment of constratint {c.pprint()} which cannot be checked using Z3")
        return False

    # We restrict our analysis to include only the subset of ScalarConstraints
    # that we can map into a Z3 model
    #
    comps = checkable(C)

    all_constraints = { x: comps[x] for x in comps }
    all_constraints[c.entityid()] = c

    vars = z3_type_map(all_constraints)

    # Compute the negation of c as a Z3 formula so that we can check
    # for the existence of a model in [[C u {c}]] that is not in [[C]]
    # (using the satisfiability check below)
    #
    negc = c.negate().to_z3(vars)

    # Populate a solver with all currently known non-redudant constraints
    # plus the negation of the selected constraint c.
    #
    s = Solver()
    for sc in comps:
        s.add(comps[sc].to_z3(vars))
    s.add(negc)

    return s.check() == unsat


# Given a set (comps, atoms) of ScalarConstraints and AtomicConstraints respectively,
# find any binding constraints in comps and use them to do constant folding within
# (comps, atoms), returning that set of constraints that is equivalent following
# one or more rounds of constant folding.
#
def fold_constants(comps, atoms):
    (bdgs, newcomps, other) = bindings_from(comps)
    blen = 0

    if len(bdgs) == 0:
        return comps, atoms

    # Otherwise, there is at least one binding to try to apply.

    while blen < len(bdgs):
        # Note that len(bdgs) is non-decreasing with each iteration
        #
        blen = len(bdgs)

        # We have not yet applied all bindings. So try to fold
        # any bound value into other constraints, simplifying
        # where possible.
        #
        for c in other:
            cprime = c.substitute(bdgs).simplify()
            if isinstance(cprime, FalseConstraint):
                # Signal that the whole set of constraints is unsatisfiable
                return {1: cprime}, {}
            if not isinstance(cprime, TrueConstraint):
                newcomps[cprime.entityid()] = cprime

        comps = newcomps
        (bdgs, newcomps, other) = bindings_from(comps)

    newatoms = {}
    for c in atoms:
        cprime = atoms[c].substitute(bdgs).simplify()
        newatoms[cprime.entityid()] = cprime

    return comps, newatoms


# Given a set (comps, atoms) of ScalarConstraints and AtomicConstraints respectively,
# find any commutative-definition constraints in comps and propagatge them within
# (comps, atoms), returning that set of constraints that is equivalent following
# one or more rounds of propagation.
#
def propagate_definitions(comps, atoms):

    (cdefs, newcomps, other) = commutative_definitions_from(comps)

    if len(cdefs) == 0:
        return comps, atoms
    for c in other:
        cprime = c.substitute(cdefs)
        newcomps[cprime.entityid()] = cprime

    newatoms = {}
    for c in atoms:
        cprime = atoms[c].substitute(cdefs)
        newatoms[cprime.entityid()] = cprime

    return newcomps, newatoms

# Given a singleton ScalarConstraint c and a set C of ScalarConstraints
# that does not include c, returns true if adding c to C makes the
# resulting set unsatisfiable
#
def incompatible_with(c, C):
    U = {v: C[v] for v in C}
    U[c.entityid()] = c
    return unsatisfiable(U)


# Let C be a set of constraints. This function attempts to find a subset C'
# of redundant constraints that can be removed without affecting the models
# of C. That is, we find:
#
#    C' subset C where [[ C ]] == [[ C \ C' ]]
#
# and returns C \ C'
#
# We compute C' by iteratively considering each c in C to decide if it
# is redundant using the following idea. It should be clear that
#
#    [[ C ]] subseteq [[ C \ {c} ]]
#
# because adding a constraint like c could only *reduce* the models
# of C \ {c}. If we can also show [[ C \ {c} ]] subseteq [[ C ]] then
# C and C \ {c} have the same models, which means that c is redundant
# and can be removed.
#
# Now suppose the contrary -- that there exists some model M where:
#
#     M |= C \ {c}
#
# but:
#
#     not( M |= C )
#
# This could only happen if c does not hold in M, i.e.:
#
#    M |= C \ {c} union { not c }
#
# The question then is, "Does there exist such an M?" Said another way,
# "Is C \ {c} union { not c } satisfiable?" Assuming that constraints
# involve only simple inequalities and integer arithmetic, we can decide
# this question using an SMT solver.
#
def reduce_constraints(C):
    # First, we restrict our analysis to include only the subset of
    # ScalarConstraints that we can map into a Z3 model
    #
    comparisons = checkable(C)

    # By default, any non-comparison constraint in C cannot be removed and so
    # must be among the reduced set of constraints
    #
    reduced = diffof(C, comparisons)

    vars = z3_type_map(comparisons)

    # Compute the negation of each comparison constraint as a Z3 formula so
    # that we can check for satisfiability below
    #
    negated = {c: comparisons[c].negate().to_z3(vars) for c in comparisons}

    # We use a workset to attempt to reduce the set of comparison constraints
    # one at a time.
    #
    working = {c: comparisons[c].to_z3(vars) for c in comparisons}

    # Those constraints that have been found to be non-redundant are stored
    # here
    #
    non_redundant = {}

    while len(working) > 0:

        # Select an arbitrary constraint c from among the workset
        k = next(iter(working))
        c = working.pop(k)

        # Populate a solver with all currently known non-redudant constraints
        # plus the negation of the selected constraint c.
        #
        s = Solver()
        for v in non_redundant.values():
            s.add(v)
        for v in working.values():
            s.add(v)
        s.add(negated[k])

        if not s.check() == unsat:
            reduced[k] = C[k]
            non_redundant[k] = c

    return reduced


# Return a dictionary of scalar variables that occur in any constraints in props
#
def scalar_occurrences(props):
    usage = {}

    for c in props.values():
        usage = c.report_propositional_variable_usage(usage)

    return usage


# Given a set of ScalarConstraints, consider the subset of those that
# are statically checkable and return True if we can prove that that
# set of constraints is not satisfiable.
#
def unsatisfiable(C):
    tocheck = checkable(C)
    if len(tocheck) == 0:
        return False

    vars = z3_type_map(tocheck)

    s = Solver()
    for c in tocheck.values():
        s.add(c.to_z3(vars))
    return s.check() == unsat


# Given a set C of propositional constraints, analyze C to return a dictionary
# of scalar variables used in C, mapping to their Z3 type.
#
def z3_type_map(C):
    vars = {}
    for c in C.values():
        vdict = c.scalar_refs()
        for v in vdict:
            if v not in vars:
                t = vdict[v].physical_typeof()
                if t.decimal():
                    vars[v] = Real(v)
                else:
                    if t.integer():
                        vars[v] = Int(v)
                    else:
                        if t.boolean():
                            vars[v] = Bool(v)
                        else:
                            raise Exception(f"Unsupported type for schema component {v}")
    return vars



import os
import sys
import re
from pathlib import Path
import traceback
from typing import Union
from colorama import Style

################################################################################
# Relative path selection
################################################################################


def find_project_root(path):
    path = Path(path).resolve()

    # Common project root indicators
    root_indicators = [
        "pyproject.toml",
        "setup.py",
        "setup.cfg",
        ".git",
        ".hg",
        ".svn",
        "requirements.txt",
        "Pipfile",
        "poetry.lock",
    ]

    current_path = path
    while current_path != current_path.parent:
        if any((current_path / indicator).exists() for indicator in root_indicators):
            return current_path
        current_path = current_path.parent

    return None


PROJECT_ROOT = find_project_root(os.getcwd())

def relative_to_project(path: Path):
    if PROJECT_ROOT and path.is_relative_to(PROJECT_ROOT):
        return path.relative_to(PROJECT_ROOT)


def relative_to_venv(path: Path):
    if hasattr(sys, "prefix"):
        venv_path = Path(sys.prefix)
        if path.is_relative_to(venv_path):
            return path.relative_to(venv_path.parent)


def relative_to_site_packages(path: Path):
    best: Union[Path, None] = None

    for site_path in sys.path:
        if "site-packages" in site_path:
            site_path = Path(site_path)
            if path.is_relative_to(site_path):
                new = path.relative_to(site_path.parent)
                if not best or len(str(best)) > len(str(new)):
                    best = new

    return best


def condense_path(path):
    resolved = Path(path).resolve()
    best: Union[Path, None] = None

    for try_option in (
        relative_to_project,
        relative_to_venv,
        relative_to_site_packages,
    ):
        option = try_option(resolved)
        if not best:
            best = option
        elif option and len(str(option)) < len(str(best)):
            best = option

    if best:
        return str(best)
    else:
        return str(path)


################################################################################
# Pretty traceback
################################################################################

def condense_traceback(exception: Exception, filter_frame_patterns = ["site-packages/_pytest", "site-packages/pluggy"]):
    tb = traceback.extract_tb(exception.__traceback__)
    max_prefix_size = 0
    prefixes: list[str] = []
    linenos: list[int] = []
    messages: list[str] = []
    for filename, lineno, name, text in tb:
        prefix = f"{condense_path(filename)} in {name}"
        exclude = False
        for pattern in filter_frame_patterns:
            if re.search(pattern, prefix):
                exclude = True
        if exclude:
            continue

        if len(prefix) > max_prefix_size:
            max_prefix_size = len(prefix)
        prefixes.append(prefix)
        linenos.append(lineno)
        messages.append(text.strip())

    return "\n".join(
        [
            f"{Style.DIM}{prefix: <{max_prefix_size}} {lineno: >4} |{Style.RESET_ALL} {msg}"
            for prefix, lineno, msg in zip(prefixes, linenos, messages)
        ]
    )



from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.relations import Relation, RelationSignature
from relationalai.early_access.dsl.core.types.standard import DateTime, Date, PositiveInteger, Decimal, String

standard_relations = {}

parse_date_name = "parse_date"
parse_date = Relation(Namespace.top, parse_date_name, RelationSignature(String, String, Date))
standard_relations[parse_date_name] = parse_date

parse_datetime_name = "parse_datetime"
parse_datetime = Relation(Namespace.top, parse_datetime_name, RelationSignature(String, String, DateTime))
standard_relations[parse_datetime_name] = parse_datetime

parse_decimal_name = "parse_decimal"
parse_decimal = Relation(Namespace.top, parse_decimal_name, RelationSignature(PositiveInteger, PositiveInteger, String, Decimal))
standard_relations[parse_decimal_name] = parse_decimal


import re
from typing import Optional

from simple_ddl_parser import DDLParser

from relationalai.early_access.dsl.core.types.standard import Integer, String, Date, Boolean, Decimal, Float, DateTime
from relationalai.early_access.dsl.core.utils import generate_stable_uuid

type_mapping = {
    "varchar": String,
    "char": String,
    "text": String,
    "int": Integer,
    "integer": Integer,
    "smallint": Integer,
    "bigint": Integer,
    "decimal": Decimal,
    "double": Float,
    "float": Float,
    "date": Date,
    "datetime": DateTime,
    "boolean": Boolean,
}


class Table:
    def __init__(self, name: Optional[str] = None, ddl_str: Optional[str] = None):
        if name and not ddl_str:
            super().__setattr__("name", name)
            super().__setattr__("columns", {})
        elif ddl_str:
            super().__setattr__("columns", {})
            parsed_dsl = self._parse_ddl(ddl_str)
            if name and name != parsed_dsl["table_name"]:
                raise ValueError(
                    f"Provided table name {name} does not match the table name in the DDL"
                )
            else:
                super().__setattr__("name", parsed_dsl["table_name"])
            for col in parsed_dsl["columns"]:
                self.columns[col["name"]] = Column(
                    col["name"], type_mapping[col["type"].lower()], self
                )
        else:
            raise ValueError("Either name or DDL must be provided")

    def _parse_ddl(self, ddl: str):
        pattern = re.compile(r"constraint_[A-Za-z_]+")
        ddl_cleaned = pattern.sub("", ddl)
        ddl_json = DDLParser(ddl_cleaned, silent=False).run()
        return ddl_json[0]

    def __setattr__(self, key: str, value: "Column") -> None:
        if key in self.__dict__:
            super().__setattr__(key, value)
        elif isinstance(value, Column):
            value.part_of = self
            self.columns[key] = value
            super().__setattr__(key, value)
        else:
           raise TypeError(
               f'Expected a Column instance for "{key}", got {type(value)}'
           )

    def __getattr__(self, key):
        if key in self.columns:
            return self.columns[key]
        else:
            raise AttributeError(f"Table {self.name} has no column named {key}")

    def pprint(self):
        return (
                f"TemporalTable: {self.name} ("
                + "\n"
                + ",\n".join(
            [f" {col.name} {col.type.pprint()}" for _, col in self.columns.items()]
        )
                + "\n)"
        )


class TemporalTable(Table):
    @property
    def temporal_col(self):
        return self._temporal_col

    def __init__(self, temporal_col_name: str, name: Optional[str] = None, ddl_str: Optional[str] = None):
        super().__init__(name, ddl_str)
        # no checks can be made here, as the columns are not yet necessarily defined
        self._temporal_col = Column(nm=temporal_col_name, type=DateTime, source=self)

    def pprint(self):
        return (
                f"Temporal Table: {self.name} ("
                + "\n"
                + ",\n".join(
            [f" {col.name} {col.type.pprint()}" for _, col in filter(lambda kv: kv[0] != '_temporal_col', self.columns.items()) ]
        )
                + "\n)"
                + f"\ntemporal_col: {self._temporal_col.name}"
        )


class Column:
    _part_of: Table

    def __init__(self, nm, type, source=None):
        self.name = nm
        self.type = type
        if source:
            self._part_of = source

    @property
    def part_of(self):
        return self._part_of

    @part_of.setter
    def part_of(self, rel):
        if self._part_of is not None and self._part_of != rel:
            raise ValueError(
                f"Column is already part of another source: {self._part_of}"
            )
        self._part_of = rel

    def pprint(self):
        return f"{self.name} {self.type.pprint()}"

    def guid(self):
        return generate_stable_uuid(f"{self._part_of.name}_{self.name}")
